import pytest

from aulos.TET12 import (
    Equal12Tuner,
    JustIntonationTuner,
    PythagoreanTuner,
)


@pytest.mark.parametrize(
    "tuner",
    [
        Equal12Tuner,
        JustIntonationTuner,
        PythagoreanTuner,
    ],
)
def test_Tuner_init(tuner):
    assert isinstance(tuner(440), tuner)



from __future__ import annotations

import typing as t
from typing import TYPE_CHECKING, cast

from aulos._core.context import inject
from aulos._core.object import AulosSchemaCollection, AulosSchemaObject
from aulos._core.pitchclass import BasePitchClass
from aulos._core.utils import classproperty, index

from ..schemas import NoteCollectionSchema, NoteSchema

if TYPE_CHECKING:
    from aulos._core.mode import BaseMode  # pragma: no cover
    from aulos._core.scale import BaseScale  # pragma: no cover
    from aulos._core.tuner import BaseTuner  # pragma: no cover


def resolve_notename_from_scale(notenumber: int, scale: BaseScale | BaseMode | None, schema: NoteSchema) -> str | None:
    if scale is not None:
        relative_pitchclass = schema.convert_notenumber_to_pitchclass(notenumber)
        relative_pitchclass = (relative_pitchclass - int(scale.key)) % schema.pitchclass.classes
        if (idx := index(scale.positions, relative_pitchclass)) is not None:
            return schema.convert_notenumber_to_notename(
                notenumber,
                scale.signatures[idx],
            )
    return None


class BaseNote[PITCHCLASS: BasePitchClass](AulosSchemaObject[NoteSchema]):
    """
    BaseNote class represents a musical note with various properties and methods for manipulation.

    This class provides the foundational structure for defining musical notes, including properties and methods
    to handle note numbers, note names, and their relationships with pitch classes and scales.
    """

    _PitchClass: t.ClassVar[type[BasePitchClass]]

    _notenumber: int
    _notenames: tuple[str | None, ...]
    _notename: str | None
    _tuner: BaseTuner | None
    _scale: BaseScale | BaseMode | None

    @inject
    def __init__(
        self,
        identify: int | str | t.Self,
        *,
        tuner: BaseTuner | None = None,
        scale: BaseScale | BaseMode | None = None,
        **kwargs: t.Any,
    ) -> None:
        super().__init__(**kwargs)

        if isinstance(identify, BaseNote):
            notenames = self.schema.convert_notenumber_to_notenames(identify.notenumber)
            notename = resolve_notename_from_scale(identify.notenumber, scale, self.schema)
            self._notenumber = identify.notenumber
            self._notenames = notenames
            self._notename = notename or identify.notename
            self._tuner = tuner or identify.tuner
            self._scale = scale or identify.scale

        elif self.is_notenumber(identify):
            notenames = self.schema.convert_notenumber_to_notenames(identify)
            notename = resolve_notename_from_scale(identify, scale, self.schema)
            self._notenumber = identify
            self._notenames = notenames
            self._notename = notename
            self._tuner = tuner
            self._scale = scale

        elif self.is_notename(identify):
            notenumber = self.schema.convert_notename_to_notenumber(identify)
            notenames = self.schema.convert_notenumber_to_notenames(notenumber)
            notename = resolve_notename_from_scale(notenumber, scale, self.schema)
            self._notenumber = notenumber
            self._notenames = notenames
            self._notename = notename or identify
            self._tuner = tuner
            self._scale = scale

        else:
            raise ValueError

    def __init_subclass__(
        cls,
        *,
        symbols_notenumber: t.Sequence[int],
        symbols_octave: t.Sequence[str],
        pitchclass: type[PITCHCLASS],
    ) -> None:
        schema = NoteSchema(
            tuple(symbols_notenumber),
            tuple(symbols_octave),
            pitchclass.schema,
        )
        super().__init_subclass__(schema=schema)
        cls._PitchClass = pitchclass

    @classproperty
    def PitchClass(self) -> type[PITCHCLASS]:  # noqa: N802
        """The type of pitch class associated with the note."""
        return cast("type[PITCHCLASS]", self._PitchClass)

    @property
    def notenumber(self) -> int:
        """Returns the note number of the note."""
        return self._notenumber

    @property
    def notenames(self) -> list[str]:
        """Returns the note names of the note."""
        return [n for n in self._notenames if n is not None]

    @property
    def notename(self) -> str | None:
        """Returns the note name of the note."""
        return self._notename

    @property
    def tuner(self) -> BaseTuner | None:
        """Returns the tuner of the note."""
        return self._tuner

    @property
    def scale(self) -> BaseScale | BaseMode | None:
        """Returns the scale of the note."""
        return self._scale

    @property
    def hz(self) -> float | None:
        """Returns the frequency of the note in hertz."""
        if self._tuner is None:
            return None
        return self._tuner.hz(self._notenumber)

    def to_pitchclass(self) -> PITCHCLASS:
        """Returns the pitch class of the note."""
        identify = (
            self.schema.convert_notename_to_pitchname(self._notename)
            if self._notename is not None
            else self.schema.convert_notenumber_to_pitchclass(self._notenumber)
        )
        return self.PitchClass(
            identify,
            scale=self._scale,
            setting=self._setting,
        )

    @classmethod
    def is_notename(cls, notename: object) -> t.TypeGuard[str]:
        """Checks if the value is a valid note name."""
        return cls.schema.is_notename(notename)

    @classmethod
    def is_notenumber(cls, notenumber: object) -> t.TypeGuard[int]:
        """Checks if the value is a valid note number."""
        return cls.schema.is_notenumber(notenumber)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, t.SupportsInt):
            return NotImplemented
        return int(self) == int(other)

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __add__(self, other: t.SupportsInt) -> t.Self:
        return self.__class__(
            int(self) + int(other),
            scale=self._scale,
            tuner=self._tuner,
            setting=self._setting,
        )

    def __sub__(self, other: t.SupportsInt) -> t.Self:
        return self.__class__(
            int(self) - int(other),
            scale=self._scale,
            tuner=self._tuner,
            setting=self._setting,
        )

    def __int__(self) -> int:
        return self._notenumber

    def __str__(self) -> str:
        return f"<Note: {self.notename or self.notenames}, scale: {self.scale}>"

    def __repr__(self) -> str:
        return f"<Note: {self.notename or self.notenames}, scale: {self.scale}>"


class BaseNoteCollection[NOTE: BaseNote](AulosSchemaCollection[NOTE, NoteCollectionSchema]):
    @inject
    def __init__(
        self,
        notes: t.Iterable[NOTE],
        **kwargs: t.Any,
    ) -> None:
        super().__init__(notes, **kwargs)

    def __init_subclass__(cls) -> None:
        schema = NoteCollectionSchema()
        super().__init_subclass__(schema=schema)

    @property
    def notes(self) -> tuple[NOTE, ...]:
        return self._objects

    def transpose(self, num: int) -> t.Self:
        return self.__class__([(note + num) for note in self._objects])

    def __eq__(self, other: object) -> bool:
        if isinstance(other, BaseNoteCollection):
            return self._objects.__eq__(other)
        return NotImplemented

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __str__(self) -> str:
        return self._objects.__str__()

    def __repr__(self) -> str:
        return self._objects.__repr__()



from aulos._core import BaseTuner

from .note import Note
from .utils import fivelimit_tuning_table, standard_tuning_table


class JustIntonationTuner(
    BaseTuner[Note],
    ratios=fivelimit_tuning_table(),
    reference_notenumber=60,
    note=Note,
):
    """
    Just Intonation Tuner (Five-limit tuning)

    * This tuner is based on the Just Intonation system, which utilizes simple
      whole-number ratios derived from the harmonic series.
    * It uses the Five-limit tuning ratios to determine the intervals between notes.
    * The system focuses on creating pure, consonant intervals, especially for
      natural harmonics.
    """


class MeantoneTuner(
    BaseTuner[Note],
    ratios=standard_tuning_table(5**0.25),
    reference_notenumber=60,
    note=Note,
):
    """
    Meantone Tuner

    * This tuner is based on the Meantone temperament, which aims to improve the
      consonance of the major third interval, often sacrificing the purity of other
      intervals like the perfect fifth.
    * Meantone tuning is typically used in historical or specific stylistic music
      applications where certain intervals are emphasized.
    """


class PythagoreanTuner(
    BaseTuner[Note],
    ratios=standard_tuning_table(1.5),
    reference_notenumber=60,
    note=Note,
):
    """
    Pythagorean Tuner

    * This tuner follows the Pythagorean tuning system, which is based on the
      harmonic series, emphasizing the pure perfect fifth (3:2 ratio).
    * The system uses the Pythagorean comma to create a series of intervals
      derived from the natural harmonic series.
    """


class Equal12Tuner(
    BaseTuner[Note],
    ratios=standard_tuning_table(2 ** (1 / 12)),
    reference_notenumber=60,
    note=Note,
):
    """
    Equal 12-Tone Tuner (12-TET)

    * This tuner uses the Equal Temperament system, where the octave is divided
      into 12 equal steps (12-TET).
    * It is the standard tuning system used in Western music, where all semitones
      are equal in frequency ratio.
    * This system allows instruments to play in all keys with equal tuning accuracy.
    """



from __future__ import annotations

import typing as t
from dataclasses import dataclass
from itertools import accumulate
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from aulos._core.utils import Positions  # pragma: no cover


@dataclass(init=False, frozen=True, slots=True)
class Intervals(t.Sequence[int]):
    _intervals: tuple[int]

    def __init__(self, iterable: t.Iterable[int]) -> None:
        object.__setattr__(self, "_intervals", tuple(iterable))

    def left(self, num: int = 1) -> Intervals:
        num %= len(self)
        return Intervals(self._intervals[num:] + self._intervals[:num])

    def right(self, num: int = 1) -> Intervals:
        num %= len(self)
        return Intervals(self._intervals[-num:] + self._intervals[:-num])

    def to_positions(self) -> Positions:
        from aulos._core.utils import Positions

        return Positions(accumulate(self._intervals[:-1]), sum(self._intervals))

    def __iter__(self) -> t.Iterator[int]:
        return self._intervals.__iter__()

    def __len__(self) -> int:
        return self._intervals.__len__()

    @t.overload
    def __getitem__(self, index: int) -> int: ...
    @t.overload
    def __getitem__(self, index: slice) -> tuple[int, ...]: ...
    def __getitem__(self, index: int | slice) -> int | tuple[int, ...]:
        return self._intervals.__getitem__(index)



import typing as t
from itertools import starmap
from typing import cast

from aulos._core.context import inject
from aulos._core.object import AulosSchemaObject
from aulos._core.pitchclass import BaseKey, BasePitchClass, PitchClassCollection
from aulos._core.utils import Intervals, Positions, classproperty

from ..schemas import ScaleSchema


class BaseScale[KEY: BaseKey, PITCHCLASS: BasePitchClass](AulosSchemaObject[ScaleSchema]):
    """
    Represents a musical scale with a specific key and pitch class.

    This class provides the foundational structure for defining musical scales, including properties and methods
    to handle intervals, positions, and key signatures. It allows for the creation and manipulation of scales
    in a theoretical context, supporting various musical keys and pitch classes.
    """

    _Key: t.ClassVar[type[BaseKey]]
    _PitchClass: t.ClassVar[type[BasePitchClass]]
    _intervals: t.ClassVar[Intervals]
    _positions: t.ClassVar[Positions]

    _key: KEY
    _signatures: tuple[int, ...]

    @inject
    def __init__(self, key: str | KEY, **kwargs: t.Any) -> None:
        super().__init__(**kwargs)

        if isinstance(key, str):
            self._key = self.Key(key, setting=self._setting)
            self._signatures = tuple(
                starmap(
                    lambda x, y: x + y,
                    zip(
                        self._key.signature,
                        self.schema.generate_scale_signatures(self._intervals),
                        strict=False,
                    ),
                ),
            )

        elif isinstance(key, BaseKey):
            self._key = key
            self._signatures = tuple(
                starmap(
                    lambda x, y: x + y,
                    zip(
                        self._key.signature,
                        self.schema.generate_scale_signatures(self._intervals),
                        strict=False,
                    ),
                ),
            )

        else:
            raise TypeError

    def __init_subclass__(
        cls,
        *,
        intervals: t.Sequence[int] | None = None,
        key: type[KEY] | None = None,
    ) -> None:
        if intervals is None or key is None:
            return
        schema = ScaleSchema(key.schema.pitchclass)
        super().__init_subclass__(schema=schema)
        cls._Key = key
        cls._PitchClass = key.PitchClass
        cls._intervals = Intervals(intervals)
        cls._positions = cls._intervals.to_positions()

    @classproperty
    def Key(self) -> type[KEY]:  # noqa: N802
        return cast("type[KEY]", self._Key)

    @classproperty
    def PitchClass(self) -> type[PITCHCLASS]:  # noqa: N802
        return cast("type[PITCHCLASS]", self._PitchClass)

    @classproperty
    def intervals(self) -> Intervals:
        return self._intervals

    @classproperty
    def positions(self) -> Positions:
        return self._positions

    @property
    def key(self) -> KEY:
        return self._key

    @property
    def signatures(self) -> tuple[int, ...]:
        return self._signatures

    @property
    def components(self) -> PitchClassCollection[PITCHCLASS]:
        components = []
        root = self.PitchClass(self._key.keyname, scale=self, setting=self.setting)
        for pos in self.positions:
            pitchclass = (root + pos).pitchclass
            note = self.PitchClass(pitchclass, scale=self, setting=self.setting)
            components.append(note)
        return PitchClassCollection(components)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BaseScale):
            return NotImplemented
        return self._intervals == other._intervals and self._key == other._key

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __str__(self) -> str:
        return f"<{self.__class__.__name__}: {self._key}>"

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self._key}>"



"""Scale
---
"""

from .bases import BaseScale
from .diatonic import DiatonicScale, NondiatonicScale
from .schemas import ScaleSchema

__all__ = [
    "BaseScale",
    "DiatonicScale",
    "NondiatonicScale",
    "ScaleSchema",
]



from __future__ import annotations

import typing as t
from typing import TYPE_CHECKING, cast

from aulos._core.context import inject
from aulos._core.note import BaseNote, NoteCollection
from aulos._core.object import AulosSchemaObject
from aulos._core.utils import Intervals, Positions, classproperty

from ..schemas import ChordSchema

if TYPE_CHECKING:
    from aulos._core.chord.quality import Quality, QualityProperty  # pragma: no cover
    from aulos._core.mode import BaseMode  # pragma: no cover
    from aulos._core.scale import BaseScale  # pragma: no cover
    from aulos._core.tuner import BaseTuner  # pragma: no cover


class BaseChord[NOTE: BaseNote](AulosSchemaObject[ChordSchema]):
    """
    Represents a musical chord, which is a combination of notes played simultaneously.

    This class extends the BaseChord and provides specific qualities and positions for various chord types,
    including triads, seventh chords, and altered chords. It allows for the representation and manipulation
    of chords in a musical context, using the Note class for individual notes.
    """

    _Note: t.ClassVar[type[BaseNote]]

    _root: NOTE
    _base: NOTE | None
    _quality: Quality
    _tuner: BaseTuner | None
    _scale: BaseScale | BaseMode | None

    @inject
    def __init__(
        self,
        identify: tuple[str, int],
        *,
        tuner: BaseTuner | None = None,
        scale: BaseScale | BaseMode | None = None,
        **kwargs: t.Any,
    ) -> None:
        super().__init__(**kwargs)

        if isinstance(identify, tuple) and isinstance(identify[0], str) and isinstance(identify[1], int):
            if (parsed := self.schema.parse(identify[0])) is not None:
                root_notename, base_notename = self.schema.convert_to_chord_notenames(
                    parsed.root,
                    parsed.base,
                    identify[1],
                )

                if base_notename is None:
                    self._root = self.Note(root_notename)
                    self._base = None
                    self._quality = parsed.quality
                    self._tuner = tuner
                    self._scale = scale

                else:
                    self._root = self.Note(root_notename)
                    self._base = self.Note(base_notename)
                    self._quality = parsed.quality.from_base(self._base.notenumber - self._root.notenumber)
                    self._tuner = tuner
                    self._scale = scale

        else:
            raise TypeError

    def __init_subclass__(cls, qualities: t.Sequence[QualityProperty], note: type[NOTE]) -> None:
        schema = ChordSchema(
            tuple(qualities),
            note.schema,
        )
        super().__init_subclass__(schema=schema)
        cls._Note = note

    @classproperty
    def Note(self) -> type[NOTE]:  # noqa: N802
        return cast("type[NOTE]", self._Note)

    @property
    def root(self) -> NOTE:
        return self._root

    @property
    def quality(self) -> Quality:
        return self._quality

    @property
    def base(self) -> NOTE | None:
        return self._base

    @property
    def tuner(self) -> BaseTuner | None:
        """Returns the tuner of the note."""
        return self._tuner

    @property
    def scale(self) -> BaseScale | BaseMode | None:
        """Returns the scale of the note."""
        return self._scale

    @property
    def intervals(self) -> Intervals:
        return self._quality.intervals

    @property
    def positions(self) -> Positions:
        return self._quality.positions

    @property
    def components(self) -> NoteCollection[NOTE]:
        components = []
        if isinstance(self._base, BaseNote) and self._quality.is_onchord():
            components.append(
                self.Note(
                    self._base,
                    tuner=self._tuner,
                    scale=self._scale,
                    setting=self._setting,
                ),
            )
        components.extend(
            self.Note(
                int(self._root) + q,
                tuner=self._tuner,
                scale=self._scale,
                setting=self._setting,
            )
            for q in self._quality
        )
        return NoteCollection(components)

    def inverse(self, num: int = 1) -> None:
        self._quality = self._quality.inverse(num)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BaseChord):
            return NotImplemented
        return self.root == other.root and self.quality == other.quality and self.base == other.base

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __str__(self) -> str:
        return f"<Chord: {self.root}{self.quality.name}{self.base}>"

    def __repr__(self) -> str:
        return f"<Chord: {self.root}{self.quality.name}{self.base}>"



"""
Aulos Library Initialization
----------------------------

This module initializes the Aulos library, a Python package designed for speech processing and
analysis from a music theory perspective. It imports essential components and modules, making
them available for use in the library.

Modules and Classes:
- TET12, TET24: Modules for handling 12-tone and 24-tone equal temperament systems.
- Setting: Core settings management for the library.

The `__all__` list defines the public API of the module, specifying which components
are accessible when the module is imported.
"""

from . import TET12, TET24
from ._core import BaseScale, BaseTuner, Setting
from ._errors import *  # noqa: F403
from ._warnings import *  # noqa: F403

__all__ = [
    "TET12",
    "TET24",
    "BaseScale",
    "BaseTuner",
    "Setting",
]



from __future__ import annotations

import typing as t
from dataclasses import dataclass

from aulos._core.utils import Intervals, Positions, index


class _OptionalQualityProperty(t.TypedDict, total=False):
    areas: tuple[str, ...]


class _RequiredQualityProperty(t.TypedDict, total=True):
    name: str
    positions: tuple[int, ...]


class QualityProperty(_RequiredQualityProperty, _OptionalQualityProperty):
    """ """


@dataclass(init=False, frozen=True, slots=True)
class Quality(t.Sequence[int]):
    name: str
    intervals: Intervals
    inversion: int
    base: int

    def __init__(
        self,
        *,
        name: str,
        positions: Positions,
        inversion: int | None = None,
        base: int | None = None,
    ) -> None:
        object.__setattr__(self, "name", name)
        object.__setattr__(self, "intervals", positions.to_intervals())
        object.__setattr__(self, "inversion", inversion or 0)
        object.__setattr__(self, "base", base or 0)

        if (base_inversion := index(self.base_candidates, base)) is not None:
            intervals = self.intervals.left(base_inversion)
            object.__setattr__(self, "intervals", intervals)
            object.__setattr__(self, "inversion", base_inversion % len(intervals))
            object.__setattr__(self, "base", 0)

    @property
    def positions(self) -> Positions:
        return self.intervals.to_positions()

    @property
    def root(self) -> int:
        return self.root_candidates[self.inversion]

    @property
    def root_candidates(self) -> tuple[int, ...]:
        return tuple(sum(self.intervals[inv:]) for inv in range(len(self.intervals), 0, -1))

    @property
    def base_candidates(self) -> tuple[int, ...]:
        return tuple(-sum(self.intervals[inv:]) for inv in range(len(self.intervals)))

    @property
    def components(self) -> tuple[int, ...]:
        return tuple(p + self.root for p in self.positions)

    def inverse(self, inversion: int) -> Quality:
        intervals = self.intervals.left(inversion)
        return Quality(
            name=self.name,
            positions=intervals.to_positions(),
            inversion=(self.inversion + inversion) % len(intervals),
            base=self.base,
        )

    def from_base(self, base: int) -> Quality:
        return Quality(name=self.name, positions=self.positions, base=base)

    def is_inverted(self) -> bool:
        return self.inversion != 0

    def is_onchord(self) -> bool:
        return self.base != 0

    def __iter__(self) -> t.Iterator[int]:
        return self.components.__iter__()

    def __len__(self) -> int:
        return self.components.__len__()

    @t.overload
    def __getitem__(self, index: int) -> int: ...
    @t.overload
    def __getitem__(self, index: slice) -> tuple[int, ...]: ...
    def __getitem__(self, index: int | slice) -> int | tuple[int, ...]:
        return self.components.__getitem__(index)



from aulos._core import BaseChord

from .note import Note


class Chord(
    BaseChord[Note],
    qualities=(
        # Triads
        {"name": "", "positions": (0, 4, 7), "areas": ("maj",)},
        {"name": "m", "positions": (0, 3, 7)},
        {"name": "sus2", "positions": (0, 2, 7)},
        {"name": "sus4", "positions": (0, 5, 7)},
        {"name": "dim", "positions": (0, 3, 6)},
        {"name": "aug", "positions": (0, 4, 8)},
        # Seventh Chords
        {"name": "6", "positions": (0, 4, 7, 9)},
        {"name": "7", "positions": (0, 4, 7, 10)},
        {"name": "M7", "positions": (0, 4, 7, 11)},
        {"name": "m6", "positions": (0, 3, 7, 9)},
        {"name": "m7", "positions": (0, 3, 7, 10)},
        {"name": "mM7", "positions": (0, 3, 7, 11)},
        {"name": "dim7", "positions": (0, 3, 6, 9)},
        {"name": "m7b5", "positions": (0, 3, 6, 10)},
        # Altered Chords
    ),
    note=Note,
):
    """
    Represents a musical chord, which is a combination of notes played simultaneously.

    This class extends the BaseChord and provides specific qualities and positions for various chord types,
    including triads, seventh chords, and altered chords. It allows for the representation and manipulation
    of chords in a musical context, using the Note class for individual notes.
    """



from __future__ import annotations

import functools
import inspect
import typing as t


def inject[**P, R](func: t.Callable[P, R]) -> t.Callable[P, R]:
    """
    Injects AulosObject dependencies from the current Context into function arguments.

    This decorator automatically provides dependencies from the current Context
    for parameters that are subclasses of AulosObject.

    Args:
        func: The function to inject dependencies into

    Returns:
        Wrapped function with injected dependencies from Context
    """

    # get function signature
    sig = inspect.signature(func)

    @functools.wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        from aulos._core.object import AulosObject

        from .context import Context

        # get context data
        injectables = Context.injectables.get({})

        for param_name in sig.parameters:
            if (
                param_name not in kwargs
                and param_name in injectables
                and isinstance(injectables[param_name], AulosObject)
            ):
                kwargs[param_name] = injectables[param_name]

        return func(*args, **kwargs)

    return wrapper



import typing as t
from dataclasses import fields, is_dataclass


def from_dict[T](cls: type[T], value: dict[str, t.Any]) -> T:
    # 1. Check if the provided class is a dataclass
    if not is_dataclass(cls.__class__):
        msg = f"The provided class {cls.__name__} is not a dataclass type."
        raise ValueError(msg)

    # 2. Convert list type in dictionary to tuple type
    # 3. Recursively convert dict to dataclass
    init_dict: dict[str, t.Any] = {}
    dfields_type = {f.name: f.type for f in fields(cls)}
    dfields_init = {f.name: f.init for f in fields(cls)}

    for k, v in value.items():
        if k in dfields_type:
            # no initialize parameter
            if not dfields_init[k]:
                continue
            # initialize parameter
            if isinstance(v, dict):
                inner_cls = dfields_type[k]
                if isinstance(inner_cls, type):
                    init_dict[k] = from_dict(inner_cls, v)
            elif isinstance(v, list):
                init_dict[k] = tuple(v)
            else:
                init_dict[k] = v

    return cls(**init_dict)



"""Core
---
"""

# utils, framework
from . import context, utils

# implementation
from .chord import BaseChord
from .mode import BaseMode
from .note import BaseNote
from .object import AulosObject, AulosSchemaObject
from .pitchclass import BaseKey, BasePitchClass
from .scale import BaseScale, DiatonicScale, NondiatonicScale
from .schema import Schema
from .setting import Setting
from .tuner import BaseTuner

__all__ = [
    "AulosObject",
    "AulosSchemaObject",
    "BaseChord",
    "BaseKey",
    "BaseMode",
    "BaseNote",
    "BasePitchClass",
    "BaseScale",
    "BaseTuner",
    "DiatonicScale",
    "NondiatonicScale",
    "Schema",
    "Setting",
    "context",
    "utils",
]



from .bases import BaseNote, BaseNoteCollection
from .note import NoteCollection
from .schemas import NoteCollectionSchema, NoteSchema

__all__ = [
    "BaseNote",
    "BaseNoteCollection",
    "NoteCollection",
    "NoteCollectionSchema",
    "NoteSchema",
]



import typing as t
from itertools import starmap
from typing import cast

from aulos._core.context import inject
from aulos._core.object import AulosSchemaObject
from aulos._core.pitchclass import BaseKey, BasePitchClass, PitchClassCollection
from aulos._core.scale import BaseScale
from aulos._core.utils import Intervals, Positions, classproperty

from ..schemas import ModeSchema


class BaseMode[KEY: BaseKey, PITCHCLASS: BasePitchClass](AulosSchemaObject[ModeSchema]):
    _Key: t.ClassVar[type[BaseKey]]
    _PitchClass: t.ClassVar[type[BasePitchClass]]
    _intervals: t.ClassVar[Intervals]
    _positions: t.ClassVar[Positions]
    _base: t.ClassVar[type[BaseScale]]

    _key: KEY
    _signatures: tuple[int, ...]

    @inject
    def __init__(self, key: str | KEY, **kwargs: t.Any) -> None:
        super().__init__(**kwargs)

        if isinstance(key, str):
            self._key = self.Key(key, setting=self._setting)
            self._signatures = tuple(
                starmap(
                    lambda x, y: x + y,
                    zip(
                        self._key.signature,
                        self.schema.scale.generate_scale_signatures(self._intervals),
                        strict=False,
                    ),
                ),
            )

        elif isinstance(key, BaseKey):
            self._key = key
            self._signatures = tuple(
                starmap(
                    lambda x, y: x + y,
                    zip(
                        self._key.signature,
                        self.schema.scale.generate_scale_signatures(self._intervals),
                        strict=False,
                    ),
                ),
            )

        else:
            raise TypeError

    def __init_subclass__(cls, *, base: type[BaseScale], shift: int, key: type[KEY]) -> None:
        schema = ModeSchema(base.schema)
        super().__init_subclass__(schema=schema)
        cls._Key = key
        cls._PitchClass = key.PitchClass
        cls._intervals = Intervals(base.intervals).left(shift)
        cls._positions = cls._intervals.to_positions()

    @classproperty
    def Key(self) -> type[KEY]:  # noqa: N802
        return cast("type[KEY]", self._Key)

    @classproperty
    def PitchClass(self) -> type[PITCHCLASS]:  # noqa: N802
        return cast("type[PITCHCLASS]", self._PitchClass)

    @classproperty
    def intervals(self) -> Intervals:
        return self._intervals

    @classproperty
    def positions(self) -> Positions:
        return self._positions

    @classproperty
    def base(self) -> type[BaseScale]:
        return self._base

    @property
    def key(self) -> KEY:
        return self._key

    @property
    def signatures(self) -> tuple[int, ...]:
        return self._signatures

    @property
    def components(self) -> PitchClassCollection[PITCHCLASS]:
        components = []
        root = self.PitchClass(self._key.keyname, scale=self, setting=self.setting)
        for pos in self.positions:
            pitchclass = (root + pos).pitchclass
            note = self.PitchClass(pitchclass, scale=self, setting=self.setting)
            components.append(note)
        return PitchClassCollection(components)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BaseScale):
            return NotImplemented
        return self._intervals == other._intervals and self._key == other._key

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __str__(self) -> str:
        return f"<{self.__class__.__name__}: {self._key}>"

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self._key}>"



from __future__ import annotations

import typing as t
from typing import TYPE_CHECKING

from aulos._core.context import inject
from aulos._core.object import AulosSchemaCollection, AulosSchemaObject
from aulos._core.pitch import PitchSchema
from aulos._core.utils import index

from ..schemas import PitchClassCollectionSchema, PitchClassSchema

if TYPE_CHECKING:
    from aulos._core.mode import BaseMode  # pragma: no cover
    from aulos._core.scale import BaseScale  # pragma: no cover


def resolve_pitchname_from_scale(
    pitchclass: int,
    scale: BaseScale | BaseMode | None,
    schema: PitchClassSchema,
) -> str | None:
    if scale is not None:
        relative_pitchclass = (pitchclass - int(scale.key)) % schema.classes
        if (idx := index(scale.positions, relative_pitchclass)) is not None:
            return schema.convert_pitchclass_to_pitchname(
                pitchclass,
                scale.signatures[idx],
            )
    return None


class BasePitchClass(AulosSchemaObject[PitchClassSchema]):
    """
    BasePitchClass represents a musical pitch class, which is
    a set of all pitches that are a whole number of octaves apart.

    This class provides the foundational structure for defining pitch classes, including properties and methods
    to handle pitch class numbers, pitch names, and their relationships within a scale.
    """

    _pitchclass: int
    _pitchnames: tuple[str | None, ...]
    _pitchname: str | None
    _scale: BaseScale | BaseMode | None

    @inject
    def __init__(
        self,
        identify: int | str | t.Self,
        *,
        scale: BaseScale | BaseMode | None = None,
        **kwargs: t.Any,
    ) -> None:
        super().__init__(**kwargs)

        if isinstance(identify, BasePitchClass):
            pitchnames = self.schema.convert_pitchclass_to_pitchnames(identify.pitchclass)
            pitchname = resolve_pitchname_from_scale(identify.pitchclass, scale, self.schema)
            self._pitchclass = identify.pitchclass
            self._pitchnames = pitchnames
            self._pitchname = pitchname or identify.pitchname
            self._scale = scale or identify.scale

        elif self.is_pitchclass(identify):
            pitchnames = self.schema.convert_pitchclass_to_pitchnames(identify)
            pitchname = resolve_pitchname_from_scale(identify, scale, self.schema)
            self._pitchclass = identify
            self._pitchnames = pitchnames
            self._pitchname = pitchname
            self._scale = scale

        elif self.is_pitchname(identify):
            pitchclass = self.schema.convert_pitchname_to_picthclass(identify)
            pitchnames = self.schema.convert_pitchclass_to_pitchnames(pitchclass)
            pitchname = resolve_pitchname_from_scale(pitchclass, scale, self.schema)
            self._pitchclass = pitchclass
            self._pitchnames = pitchnames
            self._pitchname = pitchname or identify
            self._scale = scale

        else:
            raise ValueError

    def __init_subclass__(
        cls,
        *,
        intervals: t.Sequence[int],
        symbols_pitchclass: t.Sequence[str],
        symbols_accidental: t.Sequence[str],
    ) -> None:
        schema = PitchClassSchema(
            tuple(intervals),
            tuple(symbols_pitchclass),
            tuple(symbols_accidental),
            PitchSchema(),
        )
        super().__init_subclass__(schema=schema)

    @property
    def pitchclass(self) -> int:
        """Returns the pitch class as an integer."""
        return self._pitchclass

    @property
    def pitchnames(self) -> list[str]:
        """Returns the pitch names of the pitch class."""
        return [n for n in self._pitchnames if n is not None]

    @property
    def pitchname(self) -> str | None:
        """Returns the primary pitch name of the pitch class."""
        return self._pitchname

    @property
    def scale(self) -> BaseScale | BaseMode | None:
        """Returns the scale associated with the pitch class."""
        return self._scale

    @classmethod
    def is_pitchname(cls, pitchname: object) -> t.TypeGuard[str]:
        """Checks if the value is a valid pitch name."""
        return cls.schema.is_pitchname(pitchname)

    @classmethod
    def is_pitchclass(cls, pitchclass: object) -> t.TypeGuard[int]:
        """Checks if the value is a valid pitch class."""
        return cls.schema.is_pitchclass(pitchclass)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, t.SupportsInt):
            return NotImplemented
        return int(self) == int(other)

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __add__(self, other: t.SupportsInt) -> t.Self:
        pitchclass = (int(self) + int(other)) % self.schema.classes
        return self.__class__(pitchclass, scale=self.scale, setting=self.setting)

    def __sub__(self, other: t.SupportsInt) -> t.Self:
        pitchclass = (int(self) - int(other)) % self.schema.classes
        return self.__class__(pitchclass, scale=self.scale, setting=self.setting)

    def __int__(self) -> int:
        return self._pitchclass

    def __str__(self) -> str:
        return f"<PitchClass: {self.pitchname or self.pitchnames}, scale: {self.scale}>"

    def __repr__(self) -> str:
        return f"<PitchClass: {self.pitchname or self.pitchnames}, scale: {self.scale}>"


class BasePitchClassCollection[PITCHCLASS: BasePitchClass](
    AulosSchemaCollection[PITCHCLASS, PitchClassCollectionSchema],
):
    @inject
    def __init__(
        self,
        pitchclasses: t.Iterable[PITCHCLASS],
        **kwargs: t.Any,
    ) -> None:
        super().__init__(pitchclasses, **kwargs)

    def __init_subclass__(cls) -> None:
        schema = PitchClassCollectionSchema()
        super().__init_subclass__(schema=schema)

    @property
    def pitchclasses(self) -> tuple[PITCHCLASS, ...]:
        return self._objects

    def transpose(self, num: int) -> t.Self:
        return self.__class__([(pitchclass + num) for pitchclass in self._objects])

    def inverse(self, num: int) -> t.Self:
        return self.__class__(list(self._objects[num:] + self._objects[:num]))

    def __eq__(self, other: object) -> bool:
        if isinstance(other, BasePitchClassCollection):
            return self._objects.__eq__(other)
        return NotImplemented

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __str__(self) -> str:
        return self._objects.__str__()

    def __repr__(self) -> str:
        return self._objects.__repr__()



import typing as t
from abc import ABCMeta, abstractmethod
from typing import cast

from .schema import Schema
from .setting import Setting
from .utils import classproperty


class AulosObject(metaclass=ABCMeta):
    """
    AulosObject is the base class for all objects in the Aulos framework.
    It provides a common interface and structure for all objects, including methods for equality checks,
    """

    _setting: Setting | None

    def __init__(self, setting: Setting | None = None) -> None:
        super().__init__()
        self._setting = setting

    @property
    def setting(self) -> Setting | None:
        """Returns the setting of the object."""
        return self._setting

    @abstractmethod
    def __eq__(self, other: object) -> bool: ...

    @abstractmethod
    def __ne__(self, other: object) -> bool: ...

    @abstractmethod
    def __str__(self) -> str: ...

    @abstractmethod
    def __repr__(self) -> str: ...


class AulosSchemaObject[S: Schema](AulosObject):
    """
    AulosSchemaObject is a base class for objects that are associated with a schema.
    It provides a common interface for schema validation and access to the schema.
    """

    _schema: t.ClassVar[Schema | None]

    def __init_subclass__(cls, *, schema: S | None = None) -> None:
        super().__init_subclass__()
        cls._schema = schema

    @classproperty
    def schema(self) -> S:
        if self._schema is None:
            msg = "unreachable error"
            raise RuntimeError(msg)
        return cast("S", self._schema)


class AulosSchemaCollection[T: AulosObject, S: Schema](
    t.Sequence[T],
    AulosSchemaObject[S],
):
    """
    AulosSchemaCollection is a base class for collections of AulosSchemaObjects.
    It provides a common interface for collections, including methods for accessing and manipulating the collection.
    """

    _objects: tuple[T, ...]

    def __init__(self, items: t.Iterable[T], **kwargs: t.Any) -> None:
        super().__init__(**kwargs)
        self._objects = tuple(items)

    def __iter__(self) -> t.Iterator[T]:
        return iter(self._objects)

    def __len__(self) -> int:
        return len(self._objects)

    @t.overload
    def __getitem__(self, index: int) -> T: ...

    @t.overload
    def __getitem__(self, index: slice) -> tuple[T, ...]: ...

    def __getitem__(self, index: int | slice) -> T | tuple[T, ...]:
        return self._objects.__getitem__(index)



from dataclasses import dataclass

from aulos._core.chord.quality import Quality, QualityProperty
from aulos._core.note import NoteSchema
from aulos._core.schema import Schema
from aulos._core.utils import Positions


@dataclass(frozen=True, slots=True)
class ChordComponents:
    root: str
    base: str | None
    quality: Quality


@dataclass(init=False, frozen=True, slots=True)
class ChordSchema(Schema):
    qualities: tuple[Quality, ...]
    name2quality: dict[str, Quality]
    quality2name: dict[Quality, str]
    note: NoteSchema

    def __init__(
        self,
        /,
        qualities_property: tuple[QualityProperty, ...],
        note: NoteSchema,
    ) -> None:
        super(Schema, self).__init__()

        def to_quality(quality: QualityProperty, cardinality: int) -> Quality:
            while not all(p < cardinality for p in quality["positions"]):
                cardinality *= 2
            positions = Positions(quality["positions"], cardinality)
            return Quality(
                name=quality["name"],
                positions=positions,
            )

        classes = note.pitchclass.classes
        qualities = tuple(to_quality(q, classes) for q in qualities_property)

        name2quality = {q.name: q for q in qualities}
        quality2name = {q: q.name for q in qualities}

        object.__setattr__(self, "qualities", qualities)
        object.__setattr__(self, "name2quality", name2quality)
        object.__setattr__(self, "quality2name", quality2name)
        object.__setattr__(self, "note", note)

    def validate(self) -> None:
        pass

    def parse(self, name: str) -> ChordComponents | None:
        if "/" not in name:
            root = self.note.pitchclass.find_pitchname(name)

            if root is None:
                return None

            rest = name.split(root, 1)[1]
            quality = self.name2quality[rest]
            return ChordComponents(root=root, quality=quality, base=None)
        root_quality, maybe_base = name.split("/", 1)
        base = self.note.pitchclass.find_pitchname(maybe_base)
        root = self.note.pitchclass.find_pitchname(root_quality)

        if root is None or base is None:
            return None

        rest = root_quality.split(root, 1)[1]
        quality = self.name2quality[rest]
        return ChordComponents(root=root, quality=quality, base=base)

    def convert_to_chord_notenames(
        self,
        root_pitchname: str,
        base_pitchname: str | None,
        octave: int,
    ) -> tuple[str, str | None]:
        root_notename = self.note.convert_pitchname_to_notename(root_pitchname, octave)
        if base_pitchname is None:
            return (root_notename, None)
        base_notename = self.note.find_nearest_notename(root_notename, base_pitchname, "down")
        return (root_notename, base_notename)



from dataclasses import dataclass

from aulos._core.scale import ScaleSchema
from aulos._core.schema import Schema


@dataclass(init=False, frozen=True, slots=True)
class ModeSchema(Schema):
    scale: ScaleSchema

    def __init__(self, /, scale: ScaleSchema) -> None:
        super(Schema, self).__init__()
        object.__setattr__(self, "scale", scale)

    def validate(self) -> None:
        pass



from .bases import BaseNote, BaseNoteCollection


class NoteCollection[NOTE: BaseNote](BaseNoteCollection[NOTE]):
    """
    Represents a collection of notes.
    This class extends the BaseNoteCollection and provides additional methods
    and properties to handle a collection of notes, allowing for operations
    such as adding, removing, and querying notes.
    """



from aulos._core import DiatonicScale, NondiatonicScale

from .key import Key
from .pitchclass import PitchClass


class Major(
    DiatonicScale[Key, PitchClass],
    intervals=[2, 2, 1, 2, 2, 2, 1],
    key=Key,
):
    """Major scale (Ionian mode)

    * Consisting of a specific pattern of whole and half steps: [2, 2, 1, 2, 2, 2, 1].
    * The most commonly used scale in Western music, characterized by a happy or bright sound.
    * Derived from the diatonic system, starting from the root note and ascending.
    * Frequently used as the foundation for other musical scales and modes.
    """


class Minor(
    DiatonicScale[Key, PitchClass],
    intervals=[2, 1, 2, 2, 1, 2, 2],
    key=Key,
):
    """Minor scale (Aeolian mode)

    * Consists of the following intervals: [2, 1, 2, 2, 1, 2, 2].
    * Known for its melancholic, somber, or sad sound.
    * The natural minor scale, often used in classical, rock, and pop music.
    * The third degree is lowered by a half-step compared to the major scale.
    * Forms the basis for many modal scales, including the harmonic and melodic minor scales.
    """


class HarmonicMinor(
    DiatonicScale[Key, PitchClass],
    intervals=[2, 1, 2, 2, 1, 3, 1],
    key=Key,
):
    """Harmonic minor scale

    * Consists of the following intervals: [2, 1, 2, 2, 1, 3, 1].
    * A variation of the natural minor scale with a raised 7th degree.
    * Known for its exotic, dramatic sound, often used in classical and metal music.
    * The raised 7th degree creates an augmented second between the 6th and 7th degrees.
    * Frequently used to create tension in chord progressions.
    """


class MelodicMinor(
    DiatonicScale[Key, PitchClass],
    intervals=[2, 1, 2, 2, 2, 2, 1],
    key=Key,
):
    """Melodic minor scale

    * Consists of the following intervals: [2, 1, 2, 2, 2, 2, 1].
    * In its ascending form, both the 6th and 7th degrees are raised.
    * Known for its smooth, sophisticated sound, often used in jazz and classical music.
    * The descending form typically uses the natural minor scale (Aeolian mode).
    * The raised 6th and 7th degrees help avoid awkward intervals in melodic motion.
    """


class Pentatonic(
    NondiatonicScale[Key, PitchClass],
    extensions=[[0], [0], [0], [], [0], [0], []],
    base=Major,
    key=Key,
):
    """Pentatonic scale

    * A five-note scale derived from the major scale.
    * Composed of the notes: 1, 2, 3, 5, 6 in relation to the major scale.
    * Known for its simplicity and versatility, commonly used in folk, blues, and rock music.
    * The pentatonic scale omits the 4th and 7th degrees of the major scale.
    * Works well in both major and minor contexts, creating a neutral or balanced sound.
    """


class MinorPentatonic(
    NondiatonicScale[Key, PitchClass],
    extensions=[[0], [], [0], [0], [0], [], [0]],
    base=Minor,
    key=Key,
):
    """Minor pentatonic scale

    * A five-note scale derived from the natural minor scale.
    * Composed of the notes: 1, 3, 4, 5, 7 in relation to the natural minor scale.
    * Known for its bluesy, soulful sound, commonly used in rock, blues, and jazz music.
    * Omits the 2nd and 6th degrees of the minor scale.
    * Provides a simple yet expressive framework for improvisation and melody.
    """


class Bluenote(
    NondiatonicScale[Key, PitchClass],
    extensions=[[0], [0], [-1, 0], [0], [-1, 0], [0], [-1, 0]],
    base=Major,
    key=Key,
):
    """Bluenote scale

    * A variation of the pentatonic scale with the addition of flat 5 (blue note).
    * Composed of the notes: 1, 2, 3, 4, 5b, 6, 7.
    * Characterized by its "bluesy" sound, used extensively in jazz, blues, and rock.
    * The flat 5 creates a distinct tension and release, making it effective for expressive improvisation.
    * Often associated with the "blue" feeling, evoking melancholy or sadness.
    """


class Diminish(
    NondiatonicScale[Key, PitchClass],
    extensions=[[0], [0], [-1], [0, 1], [1], [0], [0]],
    base=Major,
    key=Key,
):
    """Diminish scale

    * A symmetrical scale that alternates between whole and half steps.
    * Composed of the intervals: [2, 1, 2, 1, 2, 1, 2].
    * Known for its tension-filled, dissonant sound.
    * Frequently used in jazz and classical music for chromatic and diminished chord progressions.
    * The scale's symmetrical nature creates repeated patterns and interesting melodic possibilities.
    """


class CombDiminish(
    NondiatonicScale[Key, PitchClass],
    extensions=[[0], [-1], [-1, 0], [1], [0], [0], [-1]],
    base=Major,
    key=Key,
):
    """CombDiminish scale

    * A combination of diminished and whole-tone scale patterns.
    * Composed of alternating diminished and whole-step intervals, creating a hybrid structure.
    * Known for its mysterious, unconventional sound, often used in jazz and avant-garde music.
    * The scale's hybrid nature allows for chromatic movement and unusual chord progressions.
    """


class Wholetone(
    NondiatonicScale[Key, PitchClass],
    extensions=[[0], [0], [0], [1], [1], [1], []],
    base=Major,
    key=Key,
):
    """Whole tone scale

    * A scale composed entirely of whole steps.
    * Characterized by its ambiguous, dreamlike sound with no leading tone.
    * The whole tone scale contains only six notes in total.
    * Commonly used in impressionist music, creating a floating, ethereal atmosphere.
    * Its symmetrical structure gives it a unique sound, with no tendency to resolve to a tonic.
    """



from aulos._core import BasePitchClass


class PitchClass(
    BasePitchClass,
    intervals=(2, 2, 1, 2, 2, 2, 1),
    symbols_pitchclass=("C", "D", "E", "F", "G", "A", "B"),
    symbols_accidental=("bbb", "bb", "b", "#", "##", "###"),
):
    """
    Represents a musical pitch class, which is a set of all pitches that are a whole number of octaves apart.

    This class extends the BasePitchClass and provides specific intervals and symbols for the pitch class.
    It is used to define the basic properties of musical notes in terms of their pitch class.
    """



"""
Context Module
-------------

Provides context management and dependency injection for Aulos objects.
Includes Context class for managing application state and inject decorator
for automatic dependency resolution.
"""

from .context import Context
from .inject import inject

__all__ = ["Context", "inject"]



from aulos._core import BaseNote

from .pitchclass import PitchClass


class Note(
    BaseNote[PitchClass],
    symbols_notenumber=range(128),
    symbols_octave=(
        "<N>-1",
        "<N>0",
        "<N>1",
        "<N>2",
        "<N>3",
        "<N>4",
        "<N>5",
        "<N>6",
        "<N>7",
        "<N>8",
        "<N>9",
    ),
    pitchclass=PitchClass,
):
    """
    Represents a musical note with various properties and methods for manipulation.

    This class extends the BaseNote and associates it with a specific PitchClass.
    It provides a range of note numbers and octave symbols, allowing for the representation
    and manipulation of musical notes in a theoretical context.
    """



from abc import ABCMeta, abstractmethod


class Schema(metaclass=ABCMeta):
    def __init__(self) -> None:
        self.validate()

    @abstractmethod
    def validate(self) -> None:
        """
        Validate the schema. This method should be implemented by subclasses to perform
        any necessary validation checks on the schema.
        """
        msg = "Subclasses must implement the validate method."
        raise NotImplementedError(msg)



import pytest

from aulos.TET12 import Chord


@pytest.mark.parametrize(
    ("chord", "octave"),
    [
        ("C", 5),
        ("Cm", 5),
        ("Cm7", 5),
        ("Cm7/Bb", 5),
        ("Cm7/B", 5),
    ],
)
def test_Chord_init(chord, octave):
    assert isinstance(Chord((chord, octave)), Chord)



"""
TET12 Module Initialization
---------------------------

This module initializes the TET12 package, which is part of the Aulos library.
It provides functionalities for handling 12-tone equal temperament systems,
including musical notes, scales, and tuning systems.

Modules and Classes:
- chord: Contains the Chord class for handling musical chords.
- note: Contains classes for musical notes, keys, and pitch classes.
- scale: Provides various musical scales and modes.
- tuner: Includes different tuning systems like Equal Temperament, Just Intonation, and more.

The `__all__` list defines the public API of the module, specifying which components
are accessible when the module is imported.
"""

from .chord import Chord
from .key import Key
from .mode import (
    Aeorian,
    Aeorian_f5,
    AlteredSuperLocrian,
    Dorian,
    Dorian_f2,
    Dorian_s4,
    Ionian,
    Ionian_s5,
    Locrian,
    Locrian_n6,
    Lydian,
    Lydian_f7,
    Lydian_s2,
    Lydian_s5,
    Mixolydian,
    Mixolydian_f6,
    Mixolydian_f9,
    Phrygian,
    SuperLocrian,
)
from .note import Note
from .pitchclass import PitchClass

# Scales
# Modes
from .scale import (
    Bluenote,
    CombDiminish,
    Diminish,
    HarmonicMinor,
    Major,
    MelodicMinor,
    Minor,
    MinorPentatonic,
    Pentatonic,
    Wholetone,
)
from .tuner import Equal12Tuner, JustIntonationTuner, MeantoneTuner, PythagoreanTuner

__all__ = [
    "Aeorian",
    "Aeorian_f5",
    "AlteredSuperLocrian",
    "Bluenote",
    "Chord",
    "CombDiminish",
    "Diminish",
    "Dorian",
    "Dorian_f2",
    "Dorian_s4",
    "Equal12Tuner",
    "HarmonicMinor",
    "Ionian",
    "Ionian_s5",
    "JustIntonationTuner",
    "Key",
    "Locrian",
    "Locrian_n6",
    "Lydian",
    "Lydian_f7",
    "Lydian_s2",
    "Lydian_s5",
    "Major",
    "MeantoneTuner",
    "MelodicMinor",
    "Minor",
    "MinorPentatonic",
    "Mixolydian",
    "Mixolydian_f6",
    "Mixolydian_f9",
    "Note",
    "Pentatonic",
    "Phrygian",
    "PitchClass",
    "PythagoreanTuner",
    "SuperLocrian",
    "Wholetone",
]



"""
Utils Module
-----------

Core utility functions and tools independent of Aulos implementation.
Provides generic helpers for intervals, positions, indexing, and type operations.
"""

from .dataclass import from_dict
from .intervals import Intervals
from .positions import Positions
from .property import classproperty
from .sequence import index

__all__ = [
    "Intervals",
    "Positions",
    "classproperty",
    "from_dict",
    "index",
]



from .bases import BasePitchClass, BasePitchClassCollection


class PitchClassCollection[PITCHCLASS: BasePitchClass](BasePitchClassCollection[PITCHCLASS]):
    """
    Represents a collection of pitch classes.
    This class extends the BasePitchClassCollection and provides additional methods
    and properties to handle a collection of pitch classes, allowing for operations
    such as adding, removing, and querying pitch classes.
    """



import typing as t
from dataclasses import dataclass
from functools import cached_property
from itertools import starmap

from aulos._core.schema import Schema

from .pitchclass import PitchClassSchema


@dataclass(init=False, frozen=True, slots=True)
class KeySchema(Schema):
    accidental: int
    pitchclass: PitchClassSchema

    def __init__(self, /, accidental: int, pitchclass: PitchClassSchema) -> None:
        super(Schema, self).__init__()
        object.__setattr__(self, "accidental", accidental)
        object.__setattr__(self, "pitchclass", pitchclass)

    def validate(self) -> None:
        pass

    @cached_property
    def keynames(self) -> tuple[str, ...]:
        keynames = [
            pitchname
            for pitchname in self.pitchclass.pitchnames
            if abs(self.pitchclass.get_accidental(pitchname)) <= self.accidental
        ]
        return tuple(keynames)

    def generate_key_signatures(self, keyname: str) -> tuple[int, ...]:
        self.ensure_valid_keyname(keyname)
        std_positions = self.pitchclass.standard_positions
        key_positions = self.pitchclass.get_positions(keyname)
        return tuple(
            starmap(
                lambda x, y: x - y,
                zip(std_positions, key_positions, strict=False),
            ),
        )

    def is_keyname(self, value: object) -> t.TypeGuard[str]:
        return isinstance(value, str) and value in self.keynames

    def ensure_valid_keyname(self, keyname: str) -> None:
        if not self.is_keyname(keyname):
            raise ValueError



class ValidationError(Exception):
    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(message)

    def __str__(self) -> str:
        return f"[ValidationError]: {self.message}"


__all__ = [
    "ValidationError",
]



from aulos._core import BaseMode

from .key import Key
from .pitchclass import PitchClass
from .scale import HarmonicMinor, Major, MelodicMinor


class Ionian(
    BaseMode[Key, PitchClass],
    base=Major,
    shift=0,
    key=Key,
):
    """Ionian mode (Major scale)

    * Also known as the major scale.
    * A diatonic scale with a happy, bright, or resolved sound.
    * Composed of the intervals: [2, 2, 1, 2, 2, 2, 1].
    * The foundation of Western music harmony and melody.
    * Frequently used in a wide variety of musical genres.
    * This mode starts at the root of the major scale (no shift).
    """


class Dorian(
    BaseMode[Key, PitchClass],
    base=Major,
    shift=1,
    key=Key,
):
    """Dorian mode

    * A minor scale with a raised 6th degree.
    * Composed of the intervals: [2, 1, 2, 2, 2, 1, 2].
    * Known for its jazzy, bluesy character, balancing between major and minor.
    * The 6th degree distinguishes Dorian from the natural minor scale.
    * Often used in jazz, blues, and rock for improvisation.
    * This mode starts from the 2nd degree of the major scale.
    """


class Phrygian(
    BaseMode[Key, PitchClass],
    base=Major,
    shift=2,
    key=Key,
):
    """Phrygian mode

    * A minor scale with a lowered 2nd degree.
    * Composed of the intervals: [1, 2, 2, 2, 1, 2, 2].
    * Known for its dark, exotic, or Spanish flavor.
    * The lowered 2nd degree gives it a distinct, dissonant sound.
    * Common in flamenco music and jazz.
    * This mode starts from the 3rd degree of the major scale.
    """


class Lydian(
    BaseMode[Key, PitchClass],
    base=Major,
    shift=3,
    key=Key,
):
    """Lydian mode

    * A major scale with a raised 4th degree.
    * Composed of the intervals: [2, 2, 2, 1, 2, 2, 1].
    * Known for its bright, dreamy, or floating sound.
    * The raised 4th degree gives it a unique flavor, often used in jazz fusion and progressive rock.
    * This mode starts from the 4th degree of the major scale.
    """


class Mixolydian(
    BaseMode[Key, PitchClass],
    base=Major,
    shift=4,
    key=Key,
):
    """Mixolydian mode

    * A major scale with a lowered 7th degree.
    * Composed of the intervals: [2, 2, 1, 2, 2, 1, 2].
    * Known for its bluesy or dominant sound.
    * The lowered 7th degree gives it a less resolved sound compared to the Ionian mode.
    * Often used in rock, blues, and funk.
    * This mode starts from the 5th degree of the major scale.
    """


class Aeorian(
    BaseMode[Key, PitchClass],
    base=Major,
    shift=5,
    key=Key,
):
    """Aeolian mode (Natural minor scale)

    * A natural minor scale with no raised or lowered degrees.
    * Composed of the intervals: [2, 1, 2, 2, 1, 2, 2].
    * Known for its melancholic, somber, or sad sound.
    * The foundation for the minor scale in Western music.
    * Often used in classical, rock, and pop music.
    * This mode starts from the 6th degree of the major scale.
    """


class Locrian(
    BaseMode[Key, PitchClass],
    base=Major,
    shift=6,
    key=Key,
):
    """Locrian mode

    * A minor scale with a lowered 2nd and 5th degree.
    * Composed of the intervals: [1, 2, 2, 1, 2, 2, 2].
    * Known for its dissonant, unstable, and tense sound.
    * The lowered 5th degree gives it a diminished quality.
    * Rarely used in classical or contemporary music, but found in experimental and avant-garde genres.
    * This mode starts from the 7th degree of the major scale.
    """


class Dorian_f2(  # noqa: N801
    BaseMode[Key, PitchClass],
    base=MelodicMinor,
    shift=1,
    key=Key,
):
    """Dorian flat 2 mode (Dorian #2)

    * A modified version of the Dorian mode with a lowered 2nd degree.
    * Composed of the intervals: [1, 2, 2, 2, 2, 1, 2].
    * Known for its exotic, tense sound.
    * This mode is often used in jazz and fusion for a darker minor sound.
    * This mode starts from the 2nd degree of the melodic minor scale.
    """


class Lydian_s5(  # noqa: N801
    BaseMode[Key, PitchClass],
    base=MelodicMinor,
    shift=2,
    key=Key,
):
    """Lydian sharp 5 mode (Lydian #5)

    * A Lydian mode with a raised 5th degree.
    * Composed of the intervals: [2, 2, 2, 1, 3, 1, 2].
    * Known for its otherworldly, dreamlike quality.
    * The raised 5th degree creates an augmented chord quality.
    * Often used in jazz and experimental music for rich harmonic tension.
    * This mode starts from the 3rd degree of the melodic minor scale.
    """


class Lydian_f7(  # noqa: N801
    BaseMode[Key, PitchClass],
    base=MelodicMinor,
    shift=3,
    key=Key,
):
    """Lydian flat 7 mode (Lydian b7)

    * A Lydian mode with a lowered 7th degree.
    * Composed of the intervals: [2, 2, 2, 2, 1, 2, 1].
    * Known for its mix of major and dominant qualities.
    * Often used in jazz, fusion, and progressive rock.
    * This mode starts from the 4th degree of the melodic minor scale.
    """


class Mixolydian_f6(  # noqa: N801
    BaseMode[Key, PitchClass],
    base=MelodicMinor,
    shift=4,
    key=Key,
):
    """Mixolydian flat 6 mode (Mixolydian b6)

    * A Mixolydian mode with a lowered 6th degree.
    * Composed of the intervals: [2, 2, 1, 2, 1, 2, 2].
    * Known for its bluesy, jazzy sound.
    * The lowered 6th degree gives it a minor quality while maintaining the dominant 7th.
    * This mode is commonly used in jazz and fusion.
    * This mode starts from the 5th degree of the melodic minor scale.
    """


class Aeorian_f5(  # noqa: N801
    BaseMode[Key, PitchClass],
    base=MelodicMinor,
    shift=5,
    key=Key,
):
    """Aeolian flat 5 mode (Aeolian b5)

    * A natural minor scale with a lowered 5th degree.
    * Composed of the intervals: [2, 1, 2, 1, 2, 2, 2].
    * Known for its dark, exotic sound, often used in jazz and metal.
    * This mode creates a diminished quality by lowering the 5th degree.
    * This mode starts from the 6th degree of the melodic minor scale.
    """


class SuperLocrian(
    BaseMode[Key, PitchClass],
    base=MelodicMinor,
    shift=6,
    key=Key,
):
    """Super Locrian mode (Altered scale)

    * A scale with altered 2nd, 3rd, 4th, 5th, 6th, and 7th degrees.
    * Composed of the intervals: [1, 2, 1, 2, 1, 2, 2].
    * Known for its highly dissonant, tension-filled sound.
    * Often used in jazz over diminished or dominant chords.
    * The altered scale is a tool for creating complex harmonies.
    * This mode starts from the 7th degree of the melodic minor scale.
    """


class Locrian_n6(  # noqa: N801
    BaseMode[Key, PitchClass],
    base=HarmonicMinor,
    shift=1,
    key=Key,
):
    """Locrian natural 6 mode (Locrian #6)

    * A Locrian mode with a raised 6th degree.
    * Composed of the intervals: [1, 2, 2, 1, 2, 3, 1].
    * Known for its eerie, dissonant sound with a slight major flavor due to the raised 6th degree.
    * This mode is used in jazz and contemporary music for unusual, dissonant harmonic progressions.
    * This mode starts from the 2nd degree of the harmonic minor scale.
    """


class Ionian_s5(  # noqa: N801
    BaseMode[Key, PitchClass],
    base=HarmonicMinor,
    shift=2,
    key=Key,
):
    """Ionian sharp 5 mode (Ionian #5)

    * A major scale with a raised 5th degree.
    * Composed of the intervals: [2, 2, 1, 2, 3, 1, 1].
    * Known for its bright sound with a dramatic twist due to the sharp 5th.
    * Often used in fusion and experimental music to create harmonic tension.
    * This mode starts from the 3rd degree of the harmonic minor scale.
    """


class Dorian_s4(  # noqa: N801
    BaseMode[Key, PitchClass],
    base=HarmonicMinor,
    shift=3,
    key=Key,
):
    """Dorian sharp 4 mode (Dorian #4)

    * A Dorian mode with a raised 4th degree.
    * Composed of the intervals: [2, 1, 2, 3, 2, 1, 2].
    * Known for its tension and dark character.
    * Often used in jazz and fusion for a modern, complex sound.
    * This mode starts from the 4th degree of the harmonic minor scale.
    """


class Mixolydian_f9(  # noqa: N801
    BaseMode[Key, PitchClass],
    base=HarmonicMinor,
    shift=4,
    key=Key,
):
    """Mixolydian flat 9 mode (Phrygian Dominant)

    * A Mixolydian mode with a lowered 9th degree.
    * Composed of the intervals: [1, 2, 2, 2, 1, 2, 2].
    * Known for its exotic, Middle Eastern flavor.
    * Often used in flamenco, metal, and jazz to create tension and drama.
    * This mode starts from the 5th degree of the harmonic minor scale.
    """


class Lydian_s2(  # noqa: N801
    BaseMode[Key, PitchClass],
    base=HarmonicMinor,
    shift=5,
    key=Key,
):
    """Lydian flat 2 mode (Lydian b2)

    * A Lydian mode with a lowered 2nd degree.
    * Composed of the intervals: [1, 3, 2, 1, 2, 2, 2].
    * Known for its dramatic, exotic sound with a raised 4th and lowered 2nd degree.
    * Often used in avant-garde and fusion music to create tension and mystery.
    * This mode starts from the 6th degree of the harmonic minor scale.
    """


class AlteredSuperLocrian(
    BaseMode[Key, PitchClass],
    base=HarmonicMinor,
    shift=6,
    key=Key,
):
    """Altered Super Locrian mode (Altered scale with Super Locrian qualities)

    * A highly altered scale that is often used in jazz and fusion.
    * Composed of the intervals: [1, 2, 1, 2, 1, 2, 2].
    * Known for its extreme dissonance and tension, incorporating all possible alterations of the major scale.
    * Features a mix of both diminished and augmented intervals, contributing to its complex and unstable sound.
    * Often used in jazz over dominant chords to create maximum harmonic tension and to resolve to tonic chords.
    * This mode starts from the 7th degree of the harmonic minor scale.
    """



import pytest

from aulos.TET12 import (
    Bluenote,
    Equal12Tuner,
    HarmonicMinor,
    Ionian,
    Ionian_s5,
    JustIntonationTuner,
    Locrian,
    Lydian_f7,
    Major,
    Minor,
    Note,
    Pentatonic,
    PythagoreanTuner,
)


def test_Note_init_from_notename(data_notenames):
    for pitchname in data_notenames:
        assert isinstance(Note(pitchname), Note)


def test_Note_init_from_notenumber(data_notenumbers):
    for pitchclass in data_notenumbers:
        assert isinstance(Note(pitchclass), Note)


def test_Note_init_from_note_object(data_notenumbers):
    for notenumber in data_notenumbers:
        assert isinstance(Note(Note(notenumber)), Note)


@pytest.mark.parametrize(
    "invalid_value",
    ["", " ", "Cb-1", "G#9", -1, 128, None, [], {}],
)
def test_Note_init_from_invalid_value(invalid_value):
    with pytest.raises(ValueError):
        _ = Note(invalid_value)


def test_Note_property_get_notenumber(data_notenumbers, data_map_notename_to_notenumber):
    for notenumber in data_notenumbers:
        assert Note(notenumber).notenumber == notenumber
    for notename, notenumber in data_map_notename_to_notenumber.items():
        assert Note(notename).notenumber == notenumber


def test_Note_property_get_notename(data_notenumbers, data_notenames):
    for notenumber in data_notenumbers:
        assert Note(notenumber).notename is None
    for notename in data_notenames:
        assert Note(notename).notename == notename


def test_Note_property_get_notenames(
    data_notenumbers,
    data_notenames,
    data_map_notenumber_to_notenames,
    data_map_notename_to_notenumber,
):
    for notenumber in data_notenumbers:
        assert Note(notenumber).notenames == [
            item for item in data_map_notenumber_to_notenames[notenumber] if item is not None
        ]
    for notename in data_notenames:
        assert Note(notename).notenames == [
            item
            for item in data_map_notenumber_to_notenames[data_map_notename_to_notenumber[notename]]
            if item is not None
        ]


def test_Note_dunder_eqne(data_map_notename_to_notenumber):
    for notename, notenumber in data_map_notename_to_notenumber.items():
        assert not Note(notename) != notenumber
        assert not Note(notenumber) != notenumber
        assert not Note(notename) != Note(notenumber)
        assert not Note(notenumber) != Note(notename)


def test_PitchClass_dunder_eqne_notimplemented(data_notenumbers):
    for notenumber in data_notenumbers:
        assert not Note(notenumber) == object()


def test_Note_dunder_add(data_notenumbers):
    for notenumber1 in data_notenumbers:
        for notenumber2 in data_notenumbers:
            if (notenumber1 + notenumber2) in data_notenumbers:
                assert (Note(notenumber1) + notenumber2) == (notenumber1 + notenumber2)
            else:
                with pytest.raises(ValueError):
                    _ = Note(notenumber1) + notenumber2


def test_Note_dunder_sub(data_notenumbers):
    for notenumber1 in data_notenumbers:
        for notenumber2 in data_notenumbers:
            if (notenumber1 - notenumber2) in data_notenumbers:
                assert (Note(notenumber1) - notenumber2) == (notenumber1 - notenumber2)
            else:
                with pytest.raises(ValueError):
                    _ = Note(notenumber1) - notenumber2


def test_Note_dunder_int(data_notenumbers):
    for notenumber in data_notenumbers:
        assert int(Note(notenumber)) == notenumber


def test_Note_dunder_str(
    data_notenumbers,
    data_notenames,
    data_map_notenumber_to_notenames,
):
    for notenumber in data_notenumbers:
        notenames = [name for name in data_map_notenumber_to_notenames[notenumber] if name is not None]
        assert str(Note(notenumber)) == f"<Note: {notenames}, scale: None>"
    for notename in data_notenames:
        assert str(Note(notename)) == f"<Note: {notename}, scale: None>"


def test_Note_dunder_repr(
    data_notenumbers,
    data_notenames,
    data_map_notenumber_to_notenames,
):
    for notenumber in data_notenumbers:
        notenames = [name for name in data_map_notenumber_to_notenames[notenumber] if name is not None]
        assert repr(Note(notenumber)) == f"<Note: {notenames}, scale: None>"
    for notename in data_notenames:
        assert repr(Note(notename)) == f"<Note: {notename}, scale: None>"


@pytest.mark.parametrize(
    ("scale", "key", "notenumber", "expected"),
    [
        (Major, "C", 64, "E4"),
        (Minor, "C", 64, None),
        (HarmonicMinor, "C", 64, None),
        (Pentatonic, "C", 64, "E4"),
        (Bluenote, "C", 64, "E4"),
        (Ionian, "C", 64, "E4"),
        (Ionian_s5, "C", 64, "E4"),
        (Locrian, "C", 64, None),
        (Lydian_f7, "C", 64, "E4"),
    ],
)
def test_Note_init_with_scale(scale, key, notenumber, expected):
    pc = Note(notenumber, scale=scale(key))
    assert isinstance(pc, Note)
    assert pc.notename == expected


@pytest.mark.parametrize(
    ("tuner", "root", "notenumber", "expected"),
    [
        (JustIntonationTuner, 440, 67, 660.0),
        (PythagoreanTuner, 440, 67, 660.0),
        (Equal12Tuner, 440, 67, 659.2551138257401),
    ],
)
def test_Note_init_with_tuner(tuner, root, notenumber, expected):
    pc = Note(notenumber, tuner=tuner(root))
    assert isinstance(pc, Note)
    assert pc.hz == expected



from .bases import BaseKey, BasePitchClass
from .pitchclass import PitchClassCollection
from .schemas import KeySchema, PitchClassCollectionSchema, PitchClassSchema

__all__ = [
    "BaseKey",
    "BasePitchClass",
    "BasePitchClassCollection",
    "KeySchema",
    "PitchClassCollection",
    "PitchClassCollectionSchema",
    "PitchClassSchema",
]



import typing as t
from dataclasses import dataclass
from functools import cached_property
from itertools import chain

from aulos._core.pitch import PitchSchema
from aulos._core.schema import Schema
from aulos._core.utils import Intervals, Positions


def create_upper_sequences(
    symbols_pitchclass: tuple[str, ...],
    symbols_accidental: tuple[str, ...],
    standard_positions: Positions,
) -> list[list[str | None]]:
    accidentals = symbols_accidental[get_accidental_count(symbols_accidental) :]
    return [
        create_natural_sequence(symbols_pitchclass, (acc,), standard_positions)[0][-i:]
        + create_natural_sequence(symbols_pitchclass, (acc,), standard_positions)[0][:-i]
        for i, acc in enumerate(accidentals, start=1)
    ]


def create_lower_sequences(
    symbols_pitchclass: tuple[str, ...],
    symbols_accidental: tuple[str, ...],
    standard_positions: Positions,
) -> list[list[str | None]]:
    accidentals = symbols_accidental[: get_accidental_count(symbols_accidental)]
    return [
        create_natural_sequence(symbols_pitchclass, (acc,), standard_positions)[0][i:]
        + create_natural_sequence(symbols_pitchclass, (acc,), standard_positions)[0][:i]
        for i, acc in enumerate(reversed(accidentals), start=1)
    ]


def create_natural_sequence(
    symbols_pitchclass: tuple[str, ...],
    symbols_accidental: tuple[str, ...],
    standard_positions: Positions,
) -> list[list[str | None]]:
    return [
        [
            get_formated_pitchname(
                symbols_pitchclass[standard_positions.index(pos)],
                acc,
            )
            if pos in standard_positions
            else None
            for pos in range(standard_positions.limit)
        ]
        for acc in symbols_accidental
    ]


def get_formated_pitchname(
    symbols_pitchclass: str,
    symbols_accidental: str,
) -> str:
    if symbols_accidental.find("<P>") >= 0:
        return symbols_accidental.replace("<P>", symbols_pitchclass.upper())
    if symbols_accidental.find("<p>") >= 0:
        return symbols_accidental.replace("<p>", symbols_pitchclass.lower())
    return f"{symbols_pitchclass}{symbols_accidental}"


def get_accidental_count(symbols_accidental: tuple[str, ...]) -> int:
    return len(symbols_accidental) // 2


@dataclass(init=False, frozen=True, slots=True)
class PitchClassSchema(Schema):
    classes: int
    accidental: int
    symbols_pitchclass: tuple[str, ...]
    symbols_accidental: tuple[str, ...]
    standard_intervals: Intervals
    standard_positions: Positions
    name2class: dict[str, int]
    class2name: dict[int, tuple[str | None, ...]]
    pitch: PitchSchema

    def __init__(
        self,
        /,
        intervals: tuple[int, ...],
        symbols_pitchclass: tuple[str, ...],
        symbols_accidental: tuple[str, ...],
        pitch: PitchSchema,
    ) -> None:
        super(Schema, self).__init__()

        standard_intervals = Intervals(intervals)
        standard_positions = standard_intervals.to_positions()
        classes = standard_positions.limit
        accidental = get_accidental_count(symbols_accidental)

        accidental_natural_sequence = create_natural_sequence(
            symbols_pitchclass,
            ("",),
            standard_positions,
        )
        accidental_upper_sequences = create_upper_sequences(
            symbols_pitchclass,
            symbols_accidental,
            standard_positions,
        )
        accidental_lower_sequences = reversed(
            create_lower_sequences(
                symbols_pitchclass,
                symbols_accidental,
                standard_positions,
            ),
        )
        accidental_sequences = tuple(
            zip(
                *accidental_lower_sequences,
                *accidental_natural_sequence,
                *accidental_upper_sequences,
                strict=False,
            ),
        )
        name2class = [
            [(name, index) for name in names if name is not None] for index, names in enumerate(accidental_sequences)
        ]
        class2name = [(index, name) for index, name in enumerate(accidental_sequences)]

        object.__setattr__(self, "classes", classes)
        object.__setattr__(self, "accidental", accidental)
        object.__setattr__(self, "symbols_pitchclass", symbols_pitchclass)
        object.__setattr__(self, "symbols_accidental", symbols_accidental)
        object.__setattr__(self, "standard_intervals", standard_intervals)
        object.__setattr__(self, "standard_positions", standard_positions)
        object.__setattr__(self, "name2class", dict(chain.from_iterable(name2class)))
        object.__setattr__(self, "class2name", dict(class2name))
        object.__setattr__(self, "pitch", pitch)

    def validate(self) -> None:
        pass

    @cached_property
    def pitchnames(self) -> tuple[str, ...]:
        return tuple(self.name2class.keys())

    @cached_property
    def pitchclasses(self) -> tuple[int, ...]:
        return tuple(self.class2name.keys())

    # unstable
    def find_pitchname(self, value: str) -> str | None:
        finded = sorted(
            [pitchname for pitchname in self.pitchnames if value.find(pitchname) == 0],
            key=len,
            reverse=True,
        )
        return ([*finded, None])[0]

    def get_intervals(self, pitchname: str) -> Intervals:
        positions = self.get_positions(pitchname)
        return positions.to_intervals()

    def get_positions(self, pitchname: str) -> Positions:
        self.ensure_valid_pitchname(pitchname)
        pitchclass = self.convert_pitchname_to_picthclass(pitchname)
        accidental = self.get_accidental(pitchname)
        positions = [
            pos
            for pos in range(self.classes)
            if self.convert_pitchclass_to_pitchname((pos + pitchclass) % self.classes, accidental) is not None
        ]
        return Positions(positions, limit=self.classes)

    def get_accidental(self, pitchname: str) -> int:
        self.ensure_valid_pitchname(pitchname)
        pitchclass = self.convert_pitchname_to_picthclass(pitchname)
        pitchnames = self.convert_pitchclass_to_pitchnames(pitchclass)
        return pitchnames.index(pitchname) - self.accidental

    def convert_pitchclass_to_pitchname(
        self,
        pitchclass: int,
        accidental: int,
    ) -> str | None:
        self.ensure_valid_pitchclass(pitchclass)
        self.ensure_valid_accidental(accidental)
        return self.class2name[pitchclass][self.accidental + accidental]

    def convert_pitchclass_to_pitchnames(
        self,
        pitchclass: int,
    ) -> tuple[str | None, ...]:
        self.ensure_valid_pitchclass(pitchclass)
        return self.class2name[pitchclass]

    def convert_pitchname_to_picthclass(self, pitchname: str) -> int:
        self.ensure_valid_pitchname(pitchname)
        return self.name2class[pitchname]

    def convert_pitchclass_to_symbol(self, pitchclass: int) -> str | None:
        self.ensure_valid_pitchclass(pitchclass)
        return self.convert_pitchclass_to_pitchnames(pitchclass)[self.accidental]

    def convert_pitchname_to_symbol(self, pitchname: str) -> str:
        self.ensure_valid_pitchname(pitchname)
        accidental = self.get_accidental(pitchname)
        pitchclass = self.convert_pitchname_to_picthclass(pitchname)
        pitchclass = (pitchclass - accidental) % self.classes
        symbol = self.convert_pitchclass_to_pitchname(pitchclass, 0)
        if symbol is None:
            msg = "unreachable error"
            raise RuntimeError(msg)
        return symbol

    def is_symbol(self, value: object) -> t.TypeGuard[str]:
        return isinstance(value, str) and value in self.symbols_pitchclass

    def is_pitchname(self, value: object) -> t.TypeGuard[str]:
        return isinstance(value, str) and value in self.pitchnames

    def is_pitchclass(self, value: object) -> t.TypeGuard[int]:
        return isinstance(value, int) and value in self.pitchclasses

    def ensure_valid_pitchname(self, pitchname: str) -> None:
        if not self.is_pitchname(pitchname):
            msg = f"Invalid pitchname '{pitchname}'. Pitchname must be a valid musical note name {self.pitchnames[:3]}."
            raise ValueError(
                msg,
            )

    def ensure_valid_pitchclass(self, pitchclass: int) -> None:
        if not self.is_pitchclass(pitchclass):
            msg = (
                f"Invalid pitchclass '{pitchclass}'."
                f"Pitchclass must be an integer between {min(self.pitchclasses)} and"
                f"{max(self.pitchclasses)} inclusive."
            )
            raise ValueError(
                msg,
            )

    def ensure_valid_accidental(self, accidental: int) -> None:
        if not abs(accidental) <= self.accidental:
            msg = (
                f"Invalid accidental '{accidental}'. "
                f"Accidental must be within the range -{self.accidental} to +{self.accidental}."
            )
            raise ValueError(
                msg,
            )


@dataclass(init=False, frozen=True, slots=True)
class PitchClassCollectionSchema(Schema):
    def __init__(self) -> None:
        super(Schema, self).__init__()

    def validate(self) -> None:
        pass



from aulos._core import BaseKey

from .pitchclass import PitchClass


class Key(
    BaseKey[PitchClass],
    accidental=1,
    pitchclass=PitchClass,
):
    """
    Represents a musical key in a theoretical context.

    This class extends the BaseKey and is associated with a specific PitchClass.
    It provides functionality to define and manipulate musical keys, including handling accidentals.
    """



import typing as t


class classproperty[R: object]:  # noqa: N801
    def __init__(self, method: t.Callable[..., R] | None = None) -> None:
        self.fget = method

    def __get__(self, _: object, cls: object | type | None = None) -> R:
        if self.fget is None:
            msg = "unreadable attribute"
            raise AttributeError(msg)
        return self.fget(cls)

    def getter(self, method: t.Callable[..., R]) -> t.Self:
        self.fget = method
        return self



import typing as t
from typing import cast

from aulos._core.context import inject
from aulos._core.object import AulosSchemaObject
from aulos._core.utils import classproperty

from ..schemas import KeySchema
from .pitchclass import BasePitchClass


class BaseKey[PITCHCLASS: BasePitchClass](AulosSchemaObject[KeySchema]):
    """
    BaseKey class represents a musical key in a theoretical context.

    This class provides the foundational structure for defining and manipulating musical keys.
    It includes properties and methods to handle key names, key classes, and key signatures.
    """

    _PitchClass: t.ClassVar[type[BasePitchClass]]

    _keyname: str
    _pitchclass: int
    _signatures: tuple[int, ...]

    @inject
    def __init__(self, identify: str | t.Self, **kwargs: t.Any) -> None:
        super().__init__(**kwargs)

        if isinstance(identify, BaseKey):
            self._keyname = identify.keyname
            self._pitchclass = self.schema.pitchclass.convert_pitchname_to_picthclass(identify.keyname)
            self._signatures = self.schema.generate_key_signatures(identify.keyname)

        elif self.is_keyname(identify):
            self._keyname = identify
            self._pitchclass = self.schema.pitchclass.convert_pitchname_to_picthclass(identify)
            self._signatures = self.schema.generate_key_signatures(identify)

        else:
            raise ValueError

    def __init_subclass__(
        cls,
        *,
        accidental: int,
        pitchclass: type[BasePitchClass],
    ) -> None:
        schema = KeySchema(
            accidental,
            pitchclass.schema,
        )
        super().__init_subclass__(schema=schema)
        cls._PitchClass = pitchclass

    @classproperty
    def PitchClass(self) -> type[PITCHCLASS]:  # noqa: N802
        """The type of pitch class associated with the key."""
        return cast("type[PITCHCLASS]", self._PitchClass)

    @property
    def keyname(self) -> str:
        """Returns the name of the key."""
        return self._keyname

    @property
    def signature(self) -> tuple[int, ...]:
        """Returns the signature of the key."""
        return self._signatures

    def to_pitchclass(self) -> PITCHCLASS:
        """Returns the pitch class of the key."""
        return self.PitchClass(self._keyname, setting=self._setting)

    @classmethod
    def is_keyname(cls, value: object) -> t.TypeGuard[str]:
        """Checks if the value is a valid key name."""
        return cls.schema.is_keyname(value)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, t.SupportsInt):
            return NotImplemented
        return int(self) == int(other)

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __int__(self) -> int:
        return self._pitchclass

    def __str__(self) -> str:
        return f"<Key: {self.keyname}>"

    def __repr__(self) -> str:
        return f"<Key: {self.keyname}>"



import pytest

from aulos.TET12 import (
    Bluenote,
    HarmonicMinor,
    Ionian,
    Ionian_s5,
    Locrian,
    Lydian_f7,
    Major,
    Minor,
    Pentatonic,
    PitchClass,
)


def test_PitchClass_init_from_pitchname(data_pitchname):
    for pitchname in data_pitchname:
        assert isinstance(PitchClass(pitchname), PitchClass)


def test_PitchClass_init_from_pitchclass(data_pitchclass):
    for pitchclass in data_pitchclass:
        assert isinstance(PitchClass(pitchclass), PitchClass)


def test_PitchClass_init_from_pitchclass_object(data_pitchclass):
    for pitchclass in data_pitchclass:
        assert isinstance(PitchClass(PitchClass(pitchclass)), PitchClass)


@pytest.mark.parametrize(
    "invalid_value",
    ["", " ", "a", "g", "@", "H", -1, 12, None, [], {}],
)
def test_PitchClass_init_from_invalid_value(invalid_value):
    with pytest.raises(ValueError):
        _ = PitchClass(invalid_value)


def test_PitchClass_property_get_pitchclass(data_pitchclass, data_map_pitchname_to_pitchclass):
    for pitchclass in data_pitchclass:
        assert PitchClass(pitchclass).pitchclass == pitchclass
    for pitchname, pitchclass in data_map_pitchname_to_pitchclass.items():
        assert PitchClass(pitchname).pitchclass == pitchclass


def test_PitchClass_property_get_pitchname(data_pitchclass, data_pitchname):
    for pitchclass in data_pitchclass:
        assert PitchClass(pitchclass).pitchname is None
    for pitchname in data_pitchname:
        assert PitchClass(pitchname).pitchname == pitchname


def test_PitchClass_property_get_pitchnames(
    data_pitchclass,
    data_pitchname,
    data_map_pitchclass_to_pitchnames,
    data_map_pitchname_to_pitchclass,
):
    for pitchclass in data_pitchclass:
        assert PitchClass(pitchclass).pitchnames == [
            item for item in data_map_pitchclass_to_pitchnames[pitchclass] if item is not None
        ]
    for pitchname in data_pitchname:
        assert PitchClass(pitchname).pitchnames == [
            item
            for item in data_map_pitchclass_to_pitchnames[data_map_pitchname_to_pitchclass[pitchname]]
            if item is not None
        ]


def test_PitchClass_dunder_eqne(data_map_pitchname_to_pitchclass):
    for pitchname, pitchclass in data_map_pitchname_to_pitchclass.items():
        assert not PitchClass(pitchname) != pitchclass
        assert not PitchClass(pitchclass) != pitchclass
        assert not PitchClass(pitchname) != PitchClass(pitchclass)
        assert not PitchClass(pitchclass) != PitchClass(pitchname)


def test_PitchClass_dunder_eqne_notimplemented(data_pitchclass):
    for pitchclass in data_pitchclass:
        assert not PitchClass(pitchclass) == object()


def test_PitchClass_dunder_add(data_pitchclass):
    for pitchclass in data_pitchclass:
        for pitchclass2 in data_pitchclass:
            assert (PitchClass(pitchclass) + pitchclass2) == (pitchclass + pitchclass2) % sum(
                PitchClass.schema.standard_intervals,
            )


def test_PitchClass_dunder_sub(data_pitchclass):
    for pitchclass in data_pitchclass:
        for pitchclass2 in data_pitchclass:
            assert (PitchClass(pitchclass) - pitchclass2) == (pitchclass - pitchclass2) % sum(
                PitchClass.schema.standard_intervals,
            )


def test_PitchClass_dunder_int(data_pitchclass):
    for pitchclass in data_pitchclass:
        assert int(PitchClass(pitchclass)) == pitchclass


def test_PitchClass_dunder_str(
    data_pitchclass,
    data_pitchname,
    data_map_pitchclass_to_pitchnames,
):
    for pitchclass in data_pitchclass:
        pitchnames = [name for name in data_map_pitchclass_to_pitchnames[pitchclass] if name is not None]
        assert str(PitchClass(pitchclass)) == f"<PitchClass: {pitchnames}, scale: None>"
    for pitchname in data_pitchname:
        assert str(PitchClass(pitchname)) == f"<PitchClass: {pitchname}, scale: None>"


def test_PitchClass_dunder_repr(
    data_pitchclass,
    data_pitchname,
    data_map_pitchclass_to_pitchnames,
):
    for pitchclass in data_pitchclass:
        pitchnames = [name for name in data_map_pitchclass_to_pitchnames[pitchclass] if name is not None]
        assert repr(PitchClass(pitchclass)) == f"<PitchClass: {pitchnames}, scale: None>"
    for pitchname in data_pitchname:
        assert repr(PitchClass(pitchname)) == f"<PitchClass: {pitchname}, scale: None>"


@pytest.mark.parametrize(
    ("scale", "key", "pitchclass", "expected"),
    [
        (Major, "C", 4, "E"),
        (Minor, "C", 4, None),
        (HarmonicMinor, "C", 4, None),
        (Pentatonic, "C", 4, "E"),
        (Bluenote, "C", 4, "E"),
        (Ionian, "C", 4, "E"),
        (Ionian_s5, "C", 4, "E"),
        (Locrian, "C", 4, None),
        (Lydian_f7, "C", 4, "E"),
    ],
)
def test_PitchClass_init_with_scale(scale, key, pitchclass, expected):
    pc = PitchClass(pitchclass, scale=scale(key))
    assert isinstance(pc, PitchClass)
    assert pc.pitchname == expected



import typing as t
from dataclasses import dataclass
from functools import cached_property
from itertools import chain

from aulos._core.pitchclass import PitchClassSchema
from aulos._core.schema import Schema
from aulos._core.utils import Positions


def create_upper_sequences(
    symbols_pitchclass: tuple[str, ...],
    symbols_accidental: tuple[str, ...],
    symbols_octave: tuple[str, ...],
    standard_positions: Positions,
) -> list[list[str | None]]:
    accidentals = symbols_accidental[get_accidental_count(symbols_accidental) :]
    return [
        create_natural_sequence(symbols_pitchclass, (acc,), symbols_octave, standard_positions)[0][-i:]
        + create_natural_sequence(symbols_pitchclass, (acc,), symbols_octave, standard_positions)[0][:-i]
        for i, acc in enumerate(accidentals, start=1)
    ]


def create_lower_sequences(
    symbols_pitchclass: tuple[str, ...],
    symbols_accidental: tuple[str, ...],
    symbols_octave: tuple[str, ...],
    standard_positions: Positions,
) -> list[list[str | None]]:
    accidentals = symbols_accidental[: get_accidental_count(symbols_accidental)]
    return [
        create_natural_sequence(symbols_pitchclass, (acc,), symbols_octave, standard_positions)[0][i:]
        + create_natural_sequence(symbols_pitchclass, (acc,), symbols_octave, standard_positions)[0][:i]
        for i, acc in enumerate(reversed(accidentals), start=1)
    ]


def create_natural_sequence(
    symbols_pitchclass: tuple[str, ...],
    symbols_accidental: tuple[str, ...],
    symbols_octave: tuple[str, ...],
    standard_positions: Positions,
) -> list[list[str | None]]:
    return [
        [
            get_formated_notename(
                get_formated_pitchname(symbols_pitchclass[standard_positions.index(pos)], symbol_accidental),
                symbol_octave,
            )
            if pos in standard_positions
            else None
            for symbol_octave in symbols_octave
            for pos in range(standard_positions.limit)
        ]
        for symbol_accidental in symbols_accidental
    ]


def get_formated_pitchname(
    symbols_pitchclass: str,
    symbols_accidental: str,
) -> str:
    if symbols_accidental.find("<P>") >= 0:
        return symbols_accidental.replace("<P>", symbols_pitchclass)
    if symbols_accidental.find("<p>") >= 0:
        return symbols_accidental.replace("<p>", symbols_pitchclass)
    return f"{symbols_pitchclass}{symbols_accidental}"


def get_formated_notename(pitchname: str, symbol_octave: str) -> str:
    if symbol_octave.find("<N>") >= 0:
        return symbol_octave.replace("<N>", pitchname)
    if symbol_octave.find("<n>") >= 0:
        return symbol_octave.replace("<n>", pitchname)
    return f"{pitchname}{symbol_octave}"


def get_accidental_count(symbols_accidental: tuple[str, ...]) -> int:
    return len(symbols_accidental) // 2


@dataclass(init=False, frozen=True, slots=True)
class NoteSchema(Schema):
    symbols_notenumber: tuple[int, ...]
    symbols_octave: tuple[str, ...]
    name2number: dict[str, int]
    number2name: dict[int, tuple[str | None]]
    pitchclass: PitchClassSchema

    def __init__(
        self,
        /,
        symbols_notenumber: tuple[int, ...],
        symbols_octave: tuple[str, ...],
        pitchclass: PitchClassSchema,
    ) -> None:
        super(Schema, self).__init__()

        accidental_natural_sequences = create_natural_sequence(
            pitchclass.symbols_pitchclass,
            ("",),
            symbols_octave,
            pitchclass.standard_positions,
        )
        accidental_upper_sequences = create_upper_sequences(
            pitchclass.symbols_pitchclass,
            pitchclass.symbols_accidental,
            symbols_octave,
            pitchclass.standard_positions,
        )
        accidental_lower_sequences = reversed(
            create_lower_sequences(
                pitchclass.symbols_pitchclass,
                pitchclass.symbols_accidental,
                symbols_octave,
                pitchclass.standard_positions,
            ),
        )
        accidental_sequences = tuple(
            zip(
                *accidental_lower_sequences,
                *accidental_natural_sequences,
                *accidental_upper_sequences,
                strict=False,
            ),
        )

        name2number = [
            [(name, index) for name in names if name is not None]
            for index, names in enumerate(accidental_sequences)
            if index in symbols_notenumber
        ]
        number2name = [(index, name) for index, name in enumerate(accidental_sequences) if index in symbols_notenumber]

        object.__setattr__(self, "symbols_notenumber", symbols_notenumber)
        object.__setattr__(self, "symbols_octave", symbols_octave)
        object.__setattr__(self, "name2number", dict(chain.from_iterable(name2number)))
        object.__setattr__(self, "number2name", dict(number2name))
        object.__setattr__(self, "pitchclass", pitchclass)

    def validate(self) -> None:
        pass

    @cached_property
    def notenames(self) -> tuple[str, ...]:
        return tuple(self.name2number.keys())

    @cached_property
    def notenumbers(self) -> tuple[int, ...]:
        return tuple(self.number2name.keys())

    def find_nearest_notename(
        self,
        reference_notename: str,
        target_pitchname: str,
        direction: t.Literal["up", "down"] = "down",
    ) -> str | None:
        self.ensure_valid_notename(reference_notename)
        self.pitchclass.ensure_valid_pitchname(target_pitchname)

        if direction == "up":
            reference_notenumber = self.convert_notename_to_notenumber(reference_notename)
            target_accidental = self.pitchclass.get_accidental(target_pitchname)
            for notenumber in sorted(self.number2name.keys()):
                if notenumber > reference_notenumber:
                    candidate_notename = self.convert_notenumber_to_notename(notenumber, target_accidental)
                    if (
                        candidate_notename is not None
                        and self.convert_notename_to_pitchname(candidate_notename) == target_pitchname
                    ):
                        return candidate_notename
            return None

        if direction == "down":
            reference_notenumber = self.convert_notename_to_notenumber(reference_notename)
            target_accidental = self.pitchclass.get_accidental(target_pitchname)
            for notenumber in sorted(self.number2name.keys(), reverse=True):
                if notenumber < reference_notenumber:
                    candidate_notename = self.convert_notenumber_to_notename(notenumber, target_accidental)
                    if (
                        candidate_notename is not None
                        and self.convert_notename_to_pitchname(candidate_notename) == target_pitchname
                    ):
                        return candidate_notename
            return None
        return None

    def get_accidental(self, notename: str) -> int:
        self.ensure_valid_notename(notename)
        notenumber = self.convert_notename_to_notenumber(notename)
        notenames = self.convert_notenumber_to_notenames(notenumber)
        return notenames.index(notename) - self.pitchclass.accidental

    def get_octave(self, notenumber: int) -> int:
        self.ensure_valid_notenumber(notenumber)
        return notenumber // self.pitchclass.classes

    def convert_notenumber_to_notename(
        self,
        notenumber: int,
        accidental: int,
    ) -> str | None:
        self.ensure_valid_notenumber(notenumber)
        return self.number2name[notenumber][self.pitchclass.accidental + accidental]

    def convert_notenumber_to_notenames(
        self,
        notenumber: int,
    ) -> tuple[str | None, ...]:
        self.ensure_valid_notenumber(notenumber)
        return self.number2name[notenumber]

    def convert_notename_to_notenumber(self, notename: str) -> int:
        self.ensure_valid_notename(notename)
        return self.name2number[notename]

    def convert_notenumber_to_pitchclass(self, notenumber: int) -> int:
        self.ensure_valid_notenumber(notenumber)
        return notenumber % self.pitchclass.classes

    def convert_pitchclass_to_notenumber(self, pitchclass: int, octave: int) -> int:
        self.pitchclass.ensure_valid_pitchclass(pitchclass)
        return pitchclass + (self.pitchclass.classes * octave)

    def convert_notename_to_pitchname(self, notename: str) -> str:
        self.ensure_valid_notename(notename)
        accidental = self.get_accidental(notename)
        notenumber = self.convert_notename_to_notenumber(notename)
        pitchclass = self.convert_notenumber_to_pitchclass(notenumber)
        pitchname = self.pitchclass.convert_pitchclass_to_pitchname(
            pitchclass,
            accidental,
        )
        if pitchname is None:
            msg = "unreachable error"
            raise RuntimeError(msg)
        return pitchname

    def convert_pitchname_to_notename(self, pitchname: str, octave: int) -> str:
        self.pitchclass.ensure_valid_pitchname(pitchname)
        accidental = self.pitchclass.get_accidental(pitchname)
        pitchclass = self.pitchclass.convert_pitchname_to_picthclass(pitchname)
        notenumber = self.convert_pitchclass_to_notenumber(pitchclass, octave)
        notename = self.convert_notenumber_to_notename(notenumber, accidental)
        if notename is None:
            msg = "unreachable error"
            raise RuntimeError(msg)
        return notename

    def is_notename(self, value: object) -> t.TypeGuard[str]:
        return isinstance(value, str) and value in self.notenames

    def is_notenumber(self, value: object) -> t.TypeGuard[int]:
        return isinstance(value, int) and value in self.notenumbers

    def ensure_valid_notename(self, notename: str) -> None:
        if not self.is_notename(notename):
            msg = f"Invalid notename '{notename}'. Notename must be a valid musical note name {self.notenames[:3]}."
            raise ValueError(
                msg,
            )

    def ensure_valid_notenumber(self, notenumber: int) -> None:
        if not self.is_notenumber(notenumber):
            msg = (
                f"Invalid pitchclass '{notenumber}'."
                f"Notenumber must be an integer between {min(self.notenumbers)} and {max(self.notenumbers)} inclusive."
            )
            raise ValueError(
                msg,
            )


@dataclass(init=False, frozen=True, slots=True)
class NoteCollectionSchema(Schema):
    def __init__(self) -> None:
        super(Schema, self).__init__()

    def validate(self) -> None:
        pass



import pytest

from aulos.TET12 import (
    Bluenote,
    HarmonicMinor,
    Ionian,
    Ionian_s5,
    Key,
    Locrian,
    Lydian_f7,
    Major,
    Minor,
    Pentatonic,
)


@pytest.mark.parametrize(
    "scale",
    [
        Major,
        Minor,
        HarmonicMinor,
        Pentatonic,
        Bluenote,
        Ionian,
        Ionian_s5,
        Locrian,
        Lydian_f7,
    ],
)
def test_Scale_init_str(scale, data_keynames):
    for keyname in data_keynames:
        assert isinstance(scale(keyname), scale)


@pytest.mark.parametrize(
    "scale",
    [
        Major,
        Minor,
        HarmonicMinor,
        Pentatonic,
        Bluenote,
        Ionian,
        Ionian_s5,
        Locrian,
        Lydian_f7,
    ],
)
def test_Scale_init_key(scale, data_keynames):
    for keyname in data_keynames:
        assert isinstance(scale(Key(keyname)), scale)


@pytest.mark.parametrize(
    ("scale", "keyname", "expected_components"),
    [
        (Major, "C", ["C", "D", "E", "F", "G", "A", "B"]),
        (Minor, "C", ["C", "D", "Eb", "F", "G", "Ab", "Bb"]),
        (HarmonicMinor, "C", ["C", "D", "Eb", "F", "G", "Ab", "B"]),
        (Pentatonic, "C", ["C", "D", "E", "G", "A"]),
        (Bluenote, "C", ["C", "D", "Eb", "E", "F", "Gb", "G", "A", "Bb", "B"]),
        (Ionian, "C", ["C", "D", "E", "F", "G", "A", "B"]),
        (Ionian_s5, "C", ["C", "D", "E", "F", "G#", "A", "B"]),
        (Locrian, "C", ["C", "Db", "Eb", "F", "Gb", "Ab", "Bb"]),
        (Lydian_f7, "C", ["C", "D", "E", "F#", "G", "A", "Bb"]),
    ],
)
def test_Scale_components(scale, keyname, expected_components):
    scale_instance = scale(keyname)
    components = [pitchclass.pitchname for pitchclass in scale_instance.components]
    assert components == expected_components



import typing as t
from typing import TYPE_CHECKING

from aulos._core.utils import Intervals, Positions, classproperty

from .bases import BaseScale

# type annotaion
if TYPE_CHECKING:
    from aulos._core.pitchclass import BaseKey, BasePitchClass  # pragma: no cover


class DiatonicScale[KEY: BaseKey, PITCHCLASS: BasePitchClass](BaseScale[KEY, PITCHCLASS]):
    """
    Represents a diatonic scale, which is a musical scale consisting of seven distinct pitch classes.

    This class provides the foundational structure for creating and manipulating diatonic scales,
    allowing for the specification of intervals, key, and optional shifts in the scale's starting point.
    """


class NondiatonicScale[KEY: BaseKey, PITCHCLASS: BasePitchClass](
    BaseScale[KEY, PITCHCLASS],
):
    """
    Represents a nondiatonic scale, which extends a diatonic scale with additional intervals.

    This class allows for the creation and manipulation of scales that include extra notes beyond the
    traditional diatonic framework, providing flexibility in defining unique musical scales with
    extended harmonic possibilities.
    """

    _extensions: t.ClassVar[tuple[tuple[int, ...], ...]]
    _base: t.ClassVar[type[BaseScale]]

    def __init_subclass__(
        cls,
        *,
        extensions: t.Sequence[t.Sequence[int]],
        base: type[DiatonicScale],
        key: type[KEY],
    ) -> None:
        super().__init_subclass__(
            intervals=base.intervals,
            key=key,
        )
        cls._base = base
        cls._extensions = tuple(tuple(inner) for inner in extensions)

    @classproperty
    def intervals(self) -> Intervals:
        return self.positions.to_intervals()

    @classproperty
    def positions(self) -> Positions:
        return Positions(
            [pos + ext for pos, exts in zip(super().positions, self._extensions, strict=False) for ext in exts],
            super().positions.limit,
        )

    @property
    def signatures(self) -> tuple[int, ...]:
        return tuple(sig + ext for sig, exts in zip(super().signatures, self._extensions, strict=False) for ext in exts)



from dataclasses import dataclass

from aulos._core.note import NoteSchema
from aulos._core.pitchclass import PitchClassSchema
from aulos._core.schema import Schema


@dataclass(init=False, frozen=True, slots=True)
class TunerSchema(Schema):
    reference_notenumber: int
    note: NoteSchema
    pitchclass: PitchClassSchema

    def __init__(self, /, reference_notenumber: int, note: NoteSchema, pitchclass: PitchClassSchema) -> None:
        super(Schema, self).__init__()
        object.__setattr__(self, "reference_notenumber", reference_notenumber)
        object.__setattr__(self, "note", note)
        object.__setattr__(self, "pitchclass", pitchclass)

    def validate(self) -> None:
        pass



from dataclasses import dataclass

from aulos._core.schema import Schema

OVERTONE_RATIO = 2.0


def normalize_to_overtone_ratio(ratio: float) -> float:
    while not (1 <= ratio <= OVERTONE_RATIO):
        ratio = ratio / 2 if ratio > OVERTONE_RATIO else ratio * 2
    return ratio


@dataclass(init=False, frozen=True, slots=True)
class PitchSchema(Schema):
    def __init__(self) -> None:
        super(Schema, self).__init__()

    def validate(self) -> None:
        pass

    @classmethod
    def get_equal_tempered_ratios(cls, n: int) -> tuple[float, ...]:
        return cls.get_pitch_ratios(n, 1 / n)

    @classmethod
    def get_pythagorean_ratios(cls, n: int) -> tuple[float, ...]:
        return cls.get_pitch_ratios(n, 3 / 2)

    @classmethod
    def get_meantone_ratios(cls, n: int) -> tuple[float, ...]:
        return cls.get_pitch_ratios(n, 5 / 4)

    @classmethod
    def get_pitch_ratios(cls, n: int, ratio: float) -> tuple[float, ...]:
        return tuple(normalize_to_overtone_ratio(ratio**i) for i in range(n))



import typing as t
from typing import cast

from aulos._core.context import inject
from aulos._core.note import BaseNote
from aulos._core.object import AulosSchemaObject
from aulos._core.utils import classproperty

from ..schemas import TunerSchema


class BaseTuner[NOTE: BaseNote](AulosSchemaObject[TunerSchema]):
    """
    Represents a musical tuner that can convert note numbers to their corresponding
    frequencies in hertz (Hz) based on a specified tuning system.

    This class provides methods to handle different tuning systems, allowing for
    the conversion of musical notes into precise frequencies. It supports various tuning ratios
    and reference note numbers, making it versatile for different musical contexts.
    """

    _Note: t.ClassVar[type[BaseNote]]
    _ratios: t.ClassVar[tuple[float, ...]]

    _root: float

    @inject
    def __init__(self, root: float, **kwargs: t.Any) -> None:
        super().__init__(**kwargs)
        self._root = root

    def __init_subclass__(
        cls,
        *,
        ratios: tuple[float, ...],
        reference_notenumber: int,
        note: type[NOTE],
        **kwargs: t.Any,
    ) -> None:
        schema = TunerSchema(
            reference_notenumber,
            note.schema,
            note.schema.pitchclass,
        )
        super().__init_subclass__(schema=schema, **kwargs)
        cls._Note = note
        cls._ratios = ratios

    @classproperty
    def Note(self) -> type[NOTE]:  # noqa: N802
        """The type of note associated with the tuner."""
        return cast("type[NOTE]", self._Note)

    @classproperty
    def ratios(self) -> tuple[float, ...]:
        """The tuning ratios used to calculate frequencies."""
        return self._ratios

    @property
    def root(self) -> float:
        """Returns the root frequency of the tuner."""
        return self._root

    def hz(self, notenumber: int) -> float:
        """Converts a note number to its corresponding frequency in hertz."""
        ref = notenumber - self.schema.reference_notenumber
        octnumber = ref // self.schema.pitchclass.classes
        pitchclass = ref % self.schema.pitchclass.classes
        return self._root * (2**octnumber) * self.ratios[pitchclass]

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BaseTuner):
            return NotImplemented
        return self.ratios == other.ratios

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __str__(self) -> str:
        return f"<Tuner: {self.__class__.__name__}>"

    def __repr__(self) -> str:
        return f"<Tuner: {self.__class__.__name__}>"



import typing as t
from contextlib import ContextDecorator
from contextvars import ContextVar
from types import TracebackType

from aulos._core.object import AulosObject


class Context(ContextDecorator):
    injectables: t.ClassVar[ContextVar[dict[str, t.Any]]] = ContextVar("injectables")

    def __init__(
        self,
        **injectables: AulosObject,
    ) -> None:
        self.__injectables = self.injectables.set(injectables)

    def __enter__(self) -> t.Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        self.injectables.reset(self.__injectables)



from __future__ import annotations

import typing as t
from dataclasses import dataclass
from itertools import tee
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from aulos._core.utils import Intervals  # pragma: no cover


def diff(iterable: t.Iterable[int]) -> t.Iterator[int]:
    a, b = tee(iterable)
    next(b, None)
    return (x[1] - x[0] for x in zip(a, b, strict=False))


@dataclass(init=False, frozen=True, slots=True)
class Positions(t.Sequence[int]):
    _positions: tuple[int]
    _limit: int

    def __init__(self, iterable: t.Iterable[int], limit: int) -> None:
        positions = set(iterable)
        positions.add(0)
        object.__setattr__(self, "_positions", tuple(sorted(positions)))
        object.__setattr__(self, "_limit", limit)

    @property
    def limit(self) -> int:
        return self._limit

    def to_intervals(self) -> Intervals:
        from aulos._core.utils import Intervals

        intervals = diff([*self._positions, self._limit])
        return Intervals(intervals)

    def __iter__(self) -> t.Iterator[int]:
        return self._positions.__iter__()

    def __len__(self) -> int:
        return self._positions.__len__()

    def __contains__(self, key: object) -> bool:
        return self._positions.__contains__(key)

    @t.overload
    def __getitem__(self, index: int) -> int: ...
    @t.overload
    def __getitem__(self, index: slice) -> tuple[int, ...]: ...
    def __getitem__(self, index: int | slice) -> int | tuple[int, ...]:
        return self._positions.__getitem__(index)



from dataclasses import dataclass
from itertools import starmap

from aulos._core.pitchclass import PitchClassSchema
from aulos._core.schema import Schema
from aulos._core.utils import Intervals


@dataclass(init=False, frozen=True, slots=True)
class ScaleSchema(Schema):
    pitchclass: PitchClassSchema

    def __init__(self, /, pitchclass: PitchClassSchema) -> None:
        super(Schema, self).__init__()
        object.__setattr__(self, "pitchclass", pitchclass)

    def validate(self) -> None:
        pass

    def generate_scale_signatures(self, intervals: Intervals) -> tuple[int, ...]:
        std_positions = self.pitchclass.standard_positions
        scale_positions = intervals.to_positions()
        return tuple(
            starmap(
                lambda x, y: y - x,
                zip(std_positions, scale_positions, strict=False),
            ),
        )
